package auth

import (
	"crypto/x509"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"
	"time"

	"github.com/craigfurman/herottp"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type BearerAuthHeaderBuilder struct {
	uaaURL       string
	clientID     string
	clientSecret string
	httpClient   boshclient.HTTPClient
}

func NewBearerAuthHeaderBuilder(uaaURL, clientID, clientSecret string, disableSSLCertVerification bool, trustedCertPEM []byte) *BearerAuthHeaderBuilder {
	rootCAs := x509.NewCertPool()
	rootCAs.AppendCertsFromPEM(trustedCertPEM)
	return &BearerAuthHeaderBuilder{
		uaaURL:       uaaURL,
		clientID:     clientID,
		clientSecret: clientSecret,
		httpClient: herottp.New(herottp.Config{
			DisableTLSCertificateVerification: disableSSLCertVerification,
			RootCAs: rootCAs,
			Timeout: 30 * time.Second,
		}),
	}
}

func (hb *BearerAuthHeaderBuilder) Build() (string, error) {
	request, err := http.NewRequest(http.MethodPost, fmt.Sprintf("%s/oauth/token", hb.uaaURL), strings.NewReader("grant_type=client_credentials"))
	if err != nil {
		return "", err
	}
	request.Header.Set("Content-type", "application/x-www-form-urlencoded")
	request.SetBasicAuth(hb.clientID, hb.clientSecret)

	response, err := hb.httpClient.Do(request)
	if err != nil {
		return "", fmt.Errorf("Error reaching UAA: %s. Please ensure that properties.<broker-job>.bosh.authentication.uaa.url is correct and reachable.", err)
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		bodyBytes, err := ioutil.ReadAll(response.Body)
		if err != nil {
			bodyBytes = []byte("<error reading body>")
		}
		return "", fmt.Errorf("Error authenticating (%d): %s. Please ensure that properties.<broker-job>.bosh.authentication.uaa is correct and try again.", response.StatusCode, bodyBytes)
	}

	responseContent := map[string]interface{}{}
	err = json.NewDecoder(response.Body).Decode(&responseContent)
	if err != nil {
		return "", err
	}

	accessToken, ok := responseContent["access_token"]
	if !ok {
		return "", fmt.Errorf("no access token in grant %#v", responseContent)
	}

	return fmt.Sprintf("Bearer %s", accessToken), nil
}
