package auth

import (
	"crypto/x509"
	"encoding/pem"
	"fmt"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("BearerAuth Header Builder", func() {
	var (
		mockUAA              *mockbosh.UAA
		authorizer           boshclient.AuthHeaderBuilder
		actualClientID       = "some-username"
		actualClientSecret   = "some-password"
		suppliedClientID     string
		suppliedClientSecret string
		tokenToReturn        = "some-token"
		buildErr             error
		header               string
	)

	BeforeEach(func() {
		mockUAA = mockbosh.NewUAA(actualClientID, actualClientSecret, tokenToReturn)
	})

	AfterEach(func() {
		mockUAA.Close()
	})

	JustBeforeEach(func() {
		var certPEM []byte
		if mockUAA.TLS != nil {
			cert, err := x509.ParseCertificate(mockUAA.TLS.Certificates[0].Certificate[0])
			Expect(err).NotTo(HaveOccurred())
			certPEM = pem.EncodeToMemory(&pem.Block{Type: "CERTIFICATE", Bytes: cert.Raw})
		}
		authorizer = NewBearerAuthHeaderBuilder(mockUAA.URL, suppliedClientID, suppliedClientSecret, false, certPEM)
		header, buildErr = authorizer.Build()
	})

	Context("valid clientID and secret", func() {
		BeforeEach(func() {
			suppliedClientID = actualClientID
			suppliedClientSecret = actualClientSecret
		})

		It("succeeds", func() {
			Expect(buildErr).NotTo(HaveOccurred())
		})

		It("gets a token from UAA", func() {
			Expect(header).To(Equal(fmt.Sprintf("Bearer %s", tokenToReturn)))
		})

		Context("when the UAA server is using HTTPS with a self-signed cert", func() {
			BeforeEach(func() {
				mockUAA.Close()
				mockUAA = mockbosh.NewUAAWithTLS(actualClientID, actualClientSecret, tokenToReturn)
			})

			It("succeeds", func() {
				Expect(buildErr).NotTo(HaveOccurred())
			})
		})
	})

	Context("invalid clientID and secret", func() {
		BeforeEach(func() {
			suppliedClientID = "wrong"
			suppliedClientSecret = "wronger"
		})

		It("fails", func() {
			Expect(buildErr).To(MatchError(ContainSubstring(mockbosh.UnauthorizedError)))
			Expect(buildErr).To(MatchError(ContainSubstring(mockbosh.UnauthorizedErrorDescription)))
		})
	})

	Context("malformed response", func() {
		BeforeEach(func() {
			suppliedClientID = mockbosh.MalformedResponseUser
			suppliedClientSecret = ""
		})

		It("fails", func() {
			Expect(buildErr).To(MatchError(ContainSubstring("no access token in grant")))
		})
	})

	Context("uaa 500", func() {
		BeforeEach(func() {
			suppliedClientID = mockbosh.InternalServerErrorUser
			suppliedClientSecret = ""
		})

		It("fails", func() {
			Expect(buildErr).To(MatchError(ContainSubstring(mockbosh.InternalServerErrorMessage)))
		})
	})
})
