package boshclient_test

import (
	"errors"
	"fmt"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mockbosh"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	. "github.com/pivotal-cf/on-demand-service-broker/mockbosh"
)

func itBuildsAuthorizationHeader(err func() error) {
	It("calls the authorization header builder", func() {
		Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
	})

	Context("when the Authorization header cannot be generated", func() {
		BeforeEach(func() {
			authHeaderBuilder.BuildReturns("", errors.New("some-error"))
		})

		It("returns an error", func() {
			Expect(err()).To(MatchError(ContainSubstring("some-error")))
		})
	})
}

var _ = Describe("bosh client", func() {

	Describe("deleting bosh deployments", func() {
		var (
			taskID    int
			deleteErr error
		)

		JustBeforeEach(func() {
			taskID, deleteErr = c.DeleteDeployment("some-deployment")
		})

		Context("when bosh can delete the deployment", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					DeleteDeployment("some-deployment").RedirectsToTask(90),
				)
			})

			itBuildsAuthorizationHeader(func() error { return deleteErr })

			It("returns the bosh task ID", func() {
				Expect(taskID).To(Equal(90))
			})
		})

		Context("when bosh cannot delete the deployment", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					DeleteDeployment("some-deployment").Fails("because reasons"),
				)
			})

			itBuildsAuthorizationHeader(func() error { return deleteErr })

			It("returns an error", func() {
				Expect(deleteErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
			})
		})
	})

	Describe("getting VM info", func() {
		var (
			name = "some-deployment"

			vms    bosh.BoshVMs
			vmsErr error

			taskIDToReturn = 42
		)

		JustBeforeEach(func() {
			vms, vmsErr = c.VMs(name)
		})

		Context("when bosh starts VMs task successfully", func() {
			Context("when task state can be fetched from bosh successfully", func() {
				Context("when task output has one instance group", func() {
					BeforeEach(func() {
						director.VerifyAndMock(
							VMsForDeployment(name).RedirectsToTask(taskIDToReturn),
							Task(taskIDToReturn).ReturnsTaskWithState(boshclient.Done),
							TaskOutput(taskIDToReturn).ReturnsVMsOutput([]boshclient.BoshVMsOutput{
								{IPs: []string{"ip1", "ip2"}, InstanceGroup: "an-instance-group"},
							}),
						)
					})

					It("returns BOSH VMs", func() {
						Expect(vms).To(HaveLen(1))
						Expect(vms["an-instance-group"]).To(ConsistOf("ip1", "ip2"))
					})

					It("returns no error", func() {
						Expect(vmsErr).NotTo(HaveOccurred())
					})
				})
				Context("the output has multiple lines of the same instance group", func() {
					BeforeEach(func() {
						director.VerifyAndMock(
							VMsForDeployment(name).RedirectsToTask(taskIDToReturn),
							Task(taskIDToReturn).ReturnsTaskWithState(boshclient.Done),
							TaskOutput(taskIDToReturn).ReturnsVMsOutput([]boshclient.BoshVMsOutput{
								{IPs: []string{"ip1"}, InstanceGroup: "kafka-broker"},
								{IPs: []string{"ip2"}, InstanceGroup: "kafka-broker"},
								{IPs: []string{"ip3"}, InstanceGroup: "zookeeper"},
							}),
						)
					})

					It("groups ips by instance group name", func() {
						Expect(vms).To(HaveLen(2))
						Expect(vms).To(HaveKeyWithValue("kafka-broker", []string{"ip1", "ip2"}))
						Expect(vms).To(HaveKeyWithValue("zookeeper", []string{"ip3"}))
					})
				})

				Context("when the task is finished, but has failed", func() {
					BeforeEach(func() {
						director.VerifyAndMock(
							VMsForDeployment(name).RedirectsToTask(taskIDToReturn),
							Task(taskIDToReturn).ReturnsTaskWithState(boshclient.Error),
						)
					})

					It("returns an error", func() {
						Expect(vmsErr).To(MatchError(ContainSubstring(fmt.Sprintf("task %d failed", taskIDToReturn))))
					})
				})

				Context("when fetching task output from bosh fails", func() {
					BeforeEach(func() {
						director.VerifyAndMock(
							VMsForDeployment(name).Fails("because reasons"),
						)
					})

					It("returns an error", func() {
						Expect(vmsErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
					})
				})
			})

			Context("when fetching task state from bosh fails", func() {
				BeforeEach(func() {
					director.VerifyAndMock(
						VMsForDeployment(name).RedirectsToTask(taskIDToReturn),
						Task(taskIDToReturn).Fails("because reasons"),
					)
				})

				It("returns an error", func() {
					Expect(vmsErr).To(MatchError(ContainSubstring("expected status 200, was 500")))
				})
			})
		})

		Context("when bosh fails to start VMs task", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					VMsForDeployment(name).Fails("because reasons"),
				)
			})

			It("returns an error", func() {
				Expect(vmsErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
			})
		})
	})

	// Leaving these tests here for now, but output method will need to be made more generic at some point,
	// e.g. getting output from a failed deploy task to log it / expose it over an API.
	// At the moment, output is only used in the VMs call (above)
	Describe("getting output from a BOSH VMs task", func() {
		var (
			// name              = "some-deployment"
			taskID            = 2
			expectedVMsOutput = []boshclient.BoshVMsOutput{
				{IPs: []string{"a-nice-ip"}, InstanceGroup: "a-nice-instance-group"},
			}

			boshOutput []boshclient.BoshVMsOutput
			outputErr  error
		)

		JustBeforeEach(func() {
			boshOutput, outputErr = c.VMsOutput(taskID)
		})

		Context("when bosh succeeds", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					TaskOutput(taskID).ReturnsVMsOutput(expectedVMsOutput),
				)
			})

			itBuildsAuthorizationHeader(func() error { return outputErr })

			It("does not error", func() {
				Expect(outputErr).NotTo(HaveOccurred())
			})

			It("returns the IPs in the output", func() {
				Expect(boshOutput).To(ConsistOf(expectedVMsOutput[0]))
			})
		})

		Context("when bosh fails", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					TaskOutput(taskID).Fails("because reasons"),
				)
			})

			It("returns an error", func() {
				Expect(outputErr).To(MatchError(ContainSubstring("expected status 200, was 500.")))
			})
		})
	})

	Describe("downloading deployment manifest, as a byte array", func() {
		var (
			deploymentName   string = "some-deployment"
			manifest         []byte
			rawManifest      = []byte("a-raw-manifest")
			manifestFetchErr error
		)

		JustBeforeEach(func() {
			manifest, manifestFetchErr = c.GetManifestRaw(deploymentName)
		})

		Context("when the deployment exists", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Manifest(deploymentName).Returns(rawManifest),
				)
			})

			itBuildsAuthorizationHeader(func() error { return manifestFetchErr })

			It("returns the manifest", func() {
				Expect(manifest).To(Equal(rawManifest))
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the deployment does not exist", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Manifest(deploymentName).ReturnsManifestNotFound(),
				)
			})

			It("returns the manifest", func() {
				Expect(manifest).To(BeNil())
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})
	})

	Describe("downloading deployment manifest", func() {
		var (
			deploymentName   string = "some-deployment"
			manifest         *bosh.BoshManifest
			manifestFetchErr error
		)

		JustBeforeEach(func() {
			manifest, manifestFetchErr = c.GetManifest(deploymentName)
		})

		Context("when the deployment exists", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Manifest(deploymentName).Returns([]byte(`name: some-deployment
releases: []
stemcells: []
instance_groups: []
update:
canaries: 0
canary_watch_time: ""
update_watch_time: ""
max_in_flight: 0
`)),
				)
			})

			itBuildsAuthorizationHeader(func() error { return manifestFetchErr })

			It("returns the manifest", func() {
				Expect(manifest.Name).To(Equal(deploymentName))
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})

			Context("when the BOSH director is serving HTTPS using a self-signed certificate", func() {
				BeforeEach(func() {
					director.Close()
					director = mockbosh.NewTLS()
					director.AuthorizationHeader = expectedAuthHeader
					director.VerifyAndMock(
						Manifest(deploymentName).Returns([]byte("name: some-deployment")),
					)
				})

				It("returns no error", func() {
					Expect(manifestFetchErr).NotTo(HaveOccurred())
				})
			})
		})

		Context("when the deployment does not exist", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Manifest(deploymentName).ReturnsManifestNotFound(),
				)
			})

			It("returns the manifest", func() {
				Expect(manifest).To(BeNil())
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the manifest cannot be fetched from BOSH", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Manifest(deploymentName).Fails("because reasons"),
				)
			})

			It("returns an error", func() {
				Expect(manifestFetchErr).To(MatchError(ContainSubstring("expected status 200")))
			})
		})
	})
})
