package boshclient_test

import (
	"log"

	"gopkg.in/yaml.v2"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient/auth"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	. "github.com/pivotal-cf/on-demand-service-broker/mockbosh"
)

var _ = Describe("deploying a bosh release", func() {
	const (
		taskID = 3
	)

	var (
		deploymentManifest = bosh.BoshManifest{
			Name:           "big deployment",
			Releases:       []bosh.Release{},
			Stemcells:      []bosh.Stemcell{},
			InstanceGroups: []bosh.InstanceGroup{},
		}
		deploymentManifestBytes []byte

		returnedTaskID int
		deployErr      error
	)

	BeforeEach(func() {
		var err error
		deploymentManifestBytes, err = yaml.Marshal(deploymentManifest)
		Expect(err).NotTo(HaveOccurred())
	})

	AfterEach(func() {
		director.Close()
	})

	Context("using a structured manifest", func() {
		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Deploy().RedirectsToTask(taskID),
				)
			})
			JustBeforeEach(func() {
				returnedTaskID, deployErr = c.Deploy(deploymentManifest)
			})

			itBuildsAuthorizationHeader(func() error { return deployErr })

			It("deploys release with specified name", func() {
				Expect(deployErr).NotTo(HaveOccurred())
				Expect(director.LastCapturedManifest).To(Equal(deploymentManifestBytes))
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})
		})

		Context("when BOSH cannot deploy the release", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Deploy().RedirectsToTask(taskID).Fails("because reasons"),
				)
			})
			JustBeforeEach(func() {
				returnedTaskID, deployErr = c.Deploy(deploymentManifest)
			})

			It("returns error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("expected status 302, was 500.")))
			})
		})

		Context("when the BOSH director cannot be reached", func() {
			BeforeEach(func() {
				director.VerifyAndMock(
					Deploy().RedirectsToTask(taskID).Fails("because reasons"),
				)
			})

			It("returns error", func() {
				c = boshclient.New("http://localhost", auth.NewBasicAuthHeaderBuilder("", ""), log.New(GinkgoWriter, "[broker-unit-tests] ", log.LstdFlags), false, nil)
				_, err := c.Deploy(deploymentManifest)
				Expect(err).To(MatchError(ContainSubstring("error reaching bosh director")))
			})
		})
	})

	Context("using a raw manifest", func() {
		BeforeEach(func() {
			director.VerifyAndMock(
				Deploy().RedirectsToTask(taskID),
			)
		})
		JustBeforeEach(func() {
			deploymentManifestBytes, err := yaml.Marshal(deploymentManifest)
			Expect(err).NotTo(HaveOccurred())
			returnedTaskID, deployErr = c.DeployRaw(deploymentManifestBytes)
		})

		itBuildsAuthorizationHeader(func() error { return deployErr })

		Context("when bosh deploys the release successfully", func() {
			It("deploys release with specified name", func() {
				Expect(deployErr).NotTo(HaveOccurred())
				Expect(director.LastCapturedManifest).To(Equal(deploymentManifestBytes))
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})
		})
	})

})
