package broker_test

import (
	"errors"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Bind", func() {
	var (
		instanceID             = "a-very-impressive-instance"
		bindingID              = "I'm a binding"
		serviceDeploymentName  = deploymentName(instanceID)
		adapterBindingResponse = serviceadapter.Binding{
			Credentials:     map[string]interface{}{"foo": "bar"},
			RouteServiceURL: "route",
			SyslogDrainURL:  "syslog",
		}
		boshVms             = bosh.BoshVMs{"redis-server": []string{"an.ip"}}
		actualManifest      = []byte("valid manifest")
		arbitraryParameters = map[string]interface{}{"burns": "snrub"}

		bindResult brokerapi.Binding
		bindErr    error
	)

	BeforeEach(func() {
		boshClient.VMsReturns(boshVms, nil)
		boshClient.GetManifestRawReturns(actualManifest, nil)
		serviceAdapter.CreateBindingReturns(adapterBindingResponse, nil)
	})

	JustBeforeEach(func() {
		bindResult, bindErr = b.Bind(instanceID, bindingID, brokerapi.BindDetails{Parameters: arbitraryParameters})
	})

	It("asks bosh for VMs from a deployment named by the manifest generator", func() {
		Expect(boshClient.VMsCallCount()).To(Equal(1))
		Expect(boshClient.VMsArgsForCall(0)).To(Equal(serviceDeploymentName))
	})

	It("creates the binding using the bosh topology and admin credentials", func() {
		Expect(serviceAdapter.CreateBindingCallCount()).To(Equal(1))
		passedBindingID, passedVms, passedManifest, passedArbitraryParameters := serviceAdapter.CreateBindingArgsForCall(0)
		Expect(passedBindingID).To(Equal(bindingID))
		Expect(passedVms).To(Equal(boshVms))
		Expect(passedManifest).To(Equal(actualManifest))
		Expect(passedArbitraryParameters).To(Equal(arbitraryParameters))
	})

	It("returns the credentials, syslog drain url, and route service url produced by the service adapter", func() {
		Expect(bindResult).To(Equal(brokerapi.Binding{
			Credentials:     adapterBindingResponse.Credentials,
			SyslogDrainURL:  adapterBindingResponse.SyslogDrainURL,
			RouteServiceURL: adapterBindingResponse.RouteServiceURL,
		}))
	})

	It("does not error", func() {
		Expect(bindErr).NotTo(HaveOccurred())
	})

	Context("when bosh fails to get VMs", func() {
		BeforeEach(func() {
			boshClient.VMsReturns(nil, errors.New("oops"))
		})

		It("returns a generic error message", func() {
			Expect(bindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("oops"))
		})
	})

	Context("when manifest is not found", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, nil)
		})

		It("returns an error", func() {
			Expect(bindErr).To(HaveOccurred())
		})
	})

	Context("when broker fails to fetch manifest", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, fmt.Errorf("problem fetching manifest"))
		})

		It("returns an error", func() {
			Expect(bindErr).To(HaveOccurred())
		})
	})

	Context("when the service adapter fails to create the binding", func() {
		Context("with a generic error", func() {
			BeforeEach(func() {
				serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, errors.New("binding fail"))
			})

			It("returns a generic error message", func() {
				Expect(bindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("binding fail"))
			})
		})

		Context("with a binding already exists error", func() {
			BeforeEach(func() {
				serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, serviceadapter.NewBindingAlreadyExistsError(errors.New("binding fail")))
			})

			It("returns a binding already exists error", func() {
				Expect(bindErr).To(Equal(brokerapi.ErrBindingAlreadyExists))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("binding fail"))
			})
		})

		Context("with the app_guid not provided", func() {
			BeforeEach(func() {
				serviceAdapter.CreateBindingReturns(serviceadapter.Binding{}, serviceadapter.NewAppGuidNotProvidedError(errors.New("binding fail")))
			})

			It("returns an 'app guid not provided' error", func() {
				Expect(bindErr).To(Equal(brokerapi.ErrAppGuidNotProvided))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("binding fail"))
			})

		})
	})
})
