package broker

import (
	"encoding/json"
	"fmt"
	"log"
	"sync"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

//go:generate counterfeiter -o fake_bosh_client/fake_bosh_client.go . BoshClient
type BoshClient interface {
	DeployRaw(manifest []byte) (int, error)
	GetTasks(deploymentName string) ([]boshclient.BoshTask, error)
	VMs(instanceID string) (bosh.BoshVMs, error)
	DeleteDeployment(name string) (int, error)
	GetManifestRaw(name string) ([]byte, error)
}

//go:generate counterfeiter -o fake_instance_repository/fake_instance_repository.go . InstanceRepository
type InstanceRepository interface {
	Save(i Instance) error
	FindByID(instanceID string) (Instance, error)
	AllInstances() ([]Instance, error)
	CountByPlanID(planID string) (int, error)
	DeleteByID(instanceID string) error
	InstanceExists(instanceID string) (bool, error)
}

type OperationType string

const (
	Create  = OperationType("Create")
	Update  = OperationType("Update")
	Upgrade = OperationType("Upgrade")
	Delete  = OperationType("Delete")
)

type Instance struct {
	ID                string
	PlanID            string
	LastOperationType OperationType
}

type DeploymentNotFoundError struct {
	error
}

func NewDeploymentNotFoundError(e error) error {
	return DeploymentNotFoundError{e}
}

type DeploymentInProgressError struct {
	error
}

func NewDeploymentInProgressError(e error) error {
	return DeploymentInProgressError{e}
}

//go:generate counterfeiter -o fake_deployer/fake_deployer.go . Deployer
type Deployer interface {
	Deploy(instance Instance, planID, deploymentName string, operationType OperationType, arbitraryParams map[string]interface{}, oldManifest []byte, previousPlanID *string) (Instance, []byte, error)
}

//go:generate counterfeiter -o fake_service_adapter_client/fake_service_adapter_client.go . ServiceAdapterClient
type ServiceAdapterClient interface {
	GenerateManifest(serviceReleases serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan) ([]byte, error)
	CreateBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, arbitaryParams map[string]interface{}) (serviceadapter.Binding, error)
	DeleteBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte) error
	GenerateDashboardUrl(instanceID string, plan serviceadapter.Plan, manifest []byte) (string, error)
}

type ServiceAdapterCommandNotImplementedError struct {
	error
}

func NewServiceAdapterCommandNotImplementedError(e error) ServiceAdapterCommandNotImplementedError {
	return ServiceAdapterCommandNotImplementedError{e}
}

type Broker struct {
	boshClient     BoshClient
	instanceRepo   InstanceRepository
	serviceAdapter ServiceAdapterClient
	deployer       Deployer
	deploymentLock *sync.Mutex

	serviceOffering ServiceCatalog

	logger *log.Logger
}

func New(boshClient BoshClient, instanceRepo InstanceRepository, serviceAdapter ServiceAdapterClient,
	deployer Deployer, serviceOffering ServiceCatalog, logger *log.Logger) (*Broker, error) {

	existingPlanIDs, err := existingInstancePlanIDs(instanceRepo)
	if err != nil {
		return nil, err
	}

	for _, existingPlanID := range existingPlanIDs {
		if !containsPlanID(serviceOffering.Plans, existingPlanID) {
			return nil, fmt.Errorf("You cannot change the plan_id of a plan that has existing service instances")
		}
	}

	return &Broker{
		boshClient:     boshClient,
		instanceRepo:   instanceRepo,
		serviceAdapter: serviceAdapter,
		deployer:       deployer,
		deploymentLock: &sync.Mutex{},

		serviceOffering: serviceOffering,

		logger: logger,
	}, nil
}

func existingInstancePlanIDs(instanceRepo InstanceRepository) ([]string, error) {
	instances, err := instanceRepo.AllInstances()
	if err != nil {
		return nil, err
	}

	existingPlanIDs := []string{}
	for _, instance := range instances {
		existingPlanIDs = append(existingPlanIDs, instance.PlanID)
	}
	return existingPlanIDs, nil
}

func containsPlanID(plans []Plan, planID string) bool {
	for _, item := range plans {
		if item.ID == planID {
			return true
		}
	}
	return false
}

func convertProvisionDetailsToMap(details brokerapi.ProvisionDetails) (map[string]interface{}, error) {
	requestParams, err := convertToMap(details)
	if err != nil {
		return nil, err
	}

	if len(details.RawParameters) > 0 {
		arbitraryParams := map[string]interface{}{}
		if err := json.Unmarshal(details.RawParameters, &arbitraryParams); err != nil {
			return nil, brokerapi.ErrRawParamsInvalid
		}
		requestParams["parameters"] = arbitraryParams
	}

	return requestParams, nil
}

func convertToMap(object interface{}) (map[string]interface{}, error) {
	genericMap := map[string]interface{}{}
	data, err := json.Marshal(object)
	if err != nil {
		return nil, err
	}
	err = json.Unmarshal(data, &genericMap)

	if err != nil {
		return nil, err
	}
	return genericMap, nil
}

type DisplayableError struct {
	WrappedError error
}

func (e DisplayableError) Error() string {
	return e.WrappedError.Error()
}

func convertToDisplayableError(err error, logger *log.Logger) error {
	switch err := err.(type) {
	case nil:
		return nil
	case DisplayableError:
		return err.WrappedError
	default:
		logger.Printf("error: %s\n", err)
		return fmt.Errorf(genericErrorMessage)
	}
}

const genericErrorMessage = "An internal error has occurred"
