package broker

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type deployer struct {
	boshClient      BoshClient
	instanceRepo    InstanceRepository
	serviceAdapter  ServiceAdapterClient
	serviceOffering ServiceCatalog
	serviceStemcell serviceadapter.Stemcell
	serviceReleases serviceadapter.ServiceReleases
	logger          *log.Logger
}

func NewDeployer(
	boshClient BoshClient,
	instanceRepo InstanceRepository,
	serviceAdapter ServiceAdapterClient,
	serviceOffering ServiceCatalog,
	serviceStemcell serviceadapter.Stemcell,
	serviceReleases serviceadapter.ServiceReleases,
	logger *log.Logger,
) Deployer {
	return deployer{
		boshClient:      boshClient,
		instanceRepo:    instanceRepo,
		serviceAdapter:  serviceAdapter,
		serviceOffering: serviceOffering,
		serviceStemcell: serviceStemcell,
		serviceReleases: serviceReleases,
		logger:          logger,
	}
}

func (d deployer) Deploy(instance Instance, planID, deploymentName string, operationType OperationType, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string) (Instance, []byte, error) {
	task, err := lastOperation(d.boshClient, deploymentName)
	if err != nil && err != noTasksError {
		d.logger.Printf("error getting last operation for deployment %s: %s\n", deploymentName, err)
		return Instance{}, nil, err
	}

	if deploymentInProgress(err, task) {
		d.logger.Printf("deployment %s is still in progress: %v\n", deploymentName, task)
		return Instance{}, nil, DeploymentInProgressError{fmt.Errorf("deployment %s is still in progress", deploymentName)}
	}

	manifest, err := d.generateManifestForDeployment(deploymentName, instance, planID, requestParams, oldManifest, previousPlanID)
	if err != nil {
		d.logger.Printf("error generating manifest: %s\n", err)
		return instance, nil, err
	}

	taskID, err := d.boshClient.DeployRaw(manifest)
	if err != nil {
		return instance, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	d.logger.Printf("Bosh task ID for %s instance %s is %d\n", operationType, instance.ID, taskID)

	if err := d.updateInstance(&instance, planID, operationType); err != nil {
		return instance, nil, fmt.Errorf("error saving instance: %s", err)
	}

	return instance, manifest, nil
}

func serviceAdapterPlan(plan Plan) serviceadapter.Plan {
	abridgedPlan := serviceadapter.Plan{
		Properties:     plan.Properties,
		InstanceGroups: plan.InstanceGroups,
	}
	return abridgedPlan
}

func (d deployer) findPlans(planID string, previousPlanID *string) (serviceadapter.Plan, *serviceadapter.Plan, error) {
	plan := d.serviceOffering.FindPlanById(planID)
	if plan == nil {
		return serviceadapter.Plan{}, nil, DisplayableError{fmt.Errorf("plan %s does not exist", planID)}
	}
	var previousAdapterPlan *serviceadapter.Plan
	if previousPlanID != nil {
		previousPlan := d.serviceOffering.FindPlanById(*previousPlanID)
		if previousPlan == nil {
			return serviceadapter.Plan{}, nil, DisplayableError{fmt.Errorf("previous plan %s does not exist", *previousPlanID)}
		}
		convertedPlan := serviceAdapterPlan(*previousPlan)
		previousAdapterPlan = &convertedPlan
	}

	return serviceAdapterPlan(*plan), previousAdapterPlan, nil
}

func deploymentInProgress(err error, task boshclient.BoshTask) bool {
	return err != noTasksError && (task.State == boshclient.Queued || task.State == boshclient.Processing)
}

func (d deployer) updateInstance(instance *Instance, planID string, operationType OperationType) error {
	instance.PlanID = planID
	instance.LastOperationType = operationType
	return d.instanceRepo.Save(*instance)
}

func (d deployer) generateManifestForDeployment(deploymentName string, instance Instance, planID string, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string) ([]byte, error) {
	plan, previousPlan, err := d.findPlans(planID, previousPlanID)
	if err != nil {
		return nil, err
	}

	serviceDeployment := serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: d.serviceReleases, Stemcell: d.serviceStemcell}

	d.logger.Printf("service adapter will generate manifest for instance %s\n", instance.ID)
	return d.serviceAdapter.GenerateManifest(serviceDeployment, plan, requestParams, oldManifest, previousPlan)
}
