package broker_test

import (
	"bytes"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("Deployer", func() {

	var (
		deployer          broker.Deployer
		instance          broker.Instance
		planID            string
		previousPlanID    *string
		arbitraryParams   map[string]interface{}
		oldManifest       []byte
		deployedInstance  broker.Instance
		deployedManifest  []byte
		deployError       error
		operationType     broker.OperationType
		inMemoryLogBuffer *bytes.Buffer
		taskID            = 42

		manifest          = []byte("a-manifest")
		deploymentName    = "a-fancy-deployment"
		serviceDeployment serviceadapter.ServiceDeployment
	)

	BeforeEach(func() {
		serviceDeployment = serviceadapter.ServiceDeployment{
			Releases: serviceadapter.ServiceReleases{serviceadapter.ServiceRelease{
				Name:    "name",
				Version: "vers",
				Jobs:    []string{"a", "b"},
			}},
			Stemcell: serviceadapter.Stemcell{
				OS:      "RISC OS",
				Version: "2",
			},
		}
		inMemoryLogBuffer = bytes.NewBuffer([]byte{})
		combinedWriter := io.MultiWriter(inMemoryLogBuffer, GinkgoWriter)
		logger := log.New(combinedWriter, "[unit-tests] ", log.LstdFlags)
		deployer = broker.NewDeployer(boshClient, instanceRepo, serviceAdapter, serviceCatalog, serviceDeployment.Stemcell, serviceDeployment.Releases, logger)
		instance = broker.Instance{ID: "instance-id"}
		planID = existingPlanID
		operationType = broker.Create
		previousPlanID = stringPointer(secondPlanID)
	})

	JustBeforeEach(func() {
		deployedInstance, deployedManifest, deployError = deployer.Deploy(instance, planID, deploymentName, operationType, arbitraryParams, oldManifest, previousPlanID)
	})

	Context("deployment without previous manifest", func() {
		BeforeEach(func() {
			oldManifest = nil
			previousPlanID = nil
		})
		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				boshClient.DeployRawReturns(42, nil)
				instanceRepo.SaveReturns(nil)
				serviceAdapter.GenerateManifestReturns(manifest, nil)
				By("not having any previous tasks")
				boshClient.GetTasksReturns([]boshclient.BoshTask{}, nil)
			})

			It("checks tasks for the deployment", func() {
				Expect(boshClient.GetTasksCallCount()).To(Equal(1))
				Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
			})

			It("generates a manifest", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				actualServiceDeployment, actualPlan, actualArbParams, actualPreviousManifest, acutalPreviousPlan := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualArbParams).To(Equal(arbitraryParams))
				Expect(actualPlan).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: existingPlan.Properties}))
				Expect(actualServiceDeployment).To(Equal(serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: serviceDeployment.Releases, Stemcell: serviceDeployment.Stemcell}))
				Expect(actualPreviousManifest).To(BeNil())
				Expect(acutalPreviousPlan).To(BeNil())
			})

			It("Creates a bosh deployment using generated manifest", func() {
				Expect(boshClient.DeployRawCallCount()).To(Equal(1))
				deployedManifest := boshClient.DeployRawArgsForCall(0)
				Expect(deployedManifest).To(Equal(manifest))
			})

			It("saves the instance", func() {
				Expect(instanceRepo.SaveCallCount()).To(Equal(1))

				Expect(instanceRepo.SaveArgsForCall(0)).To(Equal(broker.Instance{
					ID:                "instance-id",
					PlanID:            existingPlanID,
					LastOperationType: operationType,
				}))
			})

			It("return the newly generated manifest", func() {
				Expect(deployedManifest).To(Equal(manifest))
			})
		})
	})
	Context("logging", func() {
		BeforeEach(func() {
			boshClient.DeployRawReturns(42, nil)
			instanceRepo.SaveReturns(nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
		})
		Context("for creating", func() {
			BeforeEach(func() {
				operationType = broker.Create
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for Create instance %s is %d", "instance-id", taskID)))
			})
		})
		Context("for updating", func() {
			BeforeEach(func() {
				operationType = broker.Update
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for Update instance %s is %d", "instance-id", taskID)))
			})
		})
		Context("for upgrading", func() {
			BeforeEach(func() {
				operationType = broker.Upgrade
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for Upgrade instance %s is %d", "instance-id", taskID)))
			})
		})
	})

	Context("deployment with previous manifest", func() {
		BeforeEach(func() {
			oldManifest = []byte("old-manifest-fetched-from-bosh")
			boshClient.DeployRawReturns(42, nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})

		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
		})

		It("Creates a bosh deployment from the generated manifest", func() {
			Expect(boshClient.DeployRawCallCount()).To(Equal(1))
			deployedManifest := boshClient.DeployRawArgsForCall(0)
			Expect(deployedManifest).To(Equal(manifest))
		})

		It("generates manifest with correct inputs", func() {
			Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
			actualServiceDeployment, planPassedToGenerate, arbParamsPassedToGenerate, previousManifestPassedToGenerate, previousPlanProperties := serviceAdapter.GenerateManifestArgsForCall(0)
			Expect(planPassedToGenerate).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: existingPlan.Properties}))
			Expect(arbParamsPassedToGenerate).To(Equal(arbitraryParams))
			Expect(previousManifestPassedToGenerate).To(Equal(oldManifest))
			Expect(actualServiceDeployment).To(Equal(serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: serviceDeployment.Releases, Stemcell: serviceDeployment.Stemcell}))
			Expect(previousPlanProperties).To(Equal(&serviceadapter.Plan{InstanceGroups: secondPlan.InstanceGroups, Properties: secondPlan.Properties}))
		})

		It("returns the deployed instance", func() {
			Expect(deployedInstance).To(Equal(broker.Instance{
				ID:                "instance-id",
				PlanID:            existingPlanID,
				LastOperationType: operationType,
			}))
		})
	})
	Context("plan isnt found", func() {
		BeforeEach(func() {
			planID = "non-existant-plan"
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})
		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("plan non-existant-plan does not exist")))
		})
	})

	Context("previous plan isnt found", func() {
		BeforeEach(func() {
			previousPlanID = stringPointer("non-existant-plan")
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})
		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("previous plan non-existant-plan does not exist")))
		})
	})

	Context("when the last bosh task for deployment is queued", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Queued}}, nil)
		})
		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(BeAssignableToTypeOf(broker.DeploymentInProgressError{}))
		})
	})

	Context("when the last bosh task for deployment is processing", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Processing}}, nil)
		})
		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(BeAssignableToTypeOf(broker.DeploymentInProgressError{}))
		})
	})

	Context("when the last bosh task for deployment fails to fetch", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns(nil, fmt.Errorf("connection error"))
		})
		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("connection error")))
		})
	})

	Context("when the manifest generator fails to generate the manifest", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
			serviceAdapter.GenerateManifestReturns(nil, errors.New("manifest fail"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("manifest fail")))
		})
	})
	Context("when the instance repository fails to save the task ID", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
			instanceRepo.SaveReturns(errors.New("failure"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("failure")))
		})
	})

	Context("when bosh fails to deploy the release", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.Done}}, nil)
			boshClient.DeployRawReturns(0, errors.New("oops"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("oops")))
		})
	})

})

func stringPointer(s string) *string {
	return &s
}
