package broker_test

import (
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("deprovisioning instances", func() {
	var (
		instanceID = "an-instance-to-be-deprovisioned"

		asyncAllowed   bool
		isAsync        brokerapi.IsAsync
		deprovisionErr error

		planId = "a_plan_id"

		deleteTaskID        = 88
		expectedInstance    broker.Instance
		expectedError       error
		instanceExists      bool
		instanceExistsError error
	)

	BeforeEach(func() {
		expectedInstance = broker.Instance{ID: instanceID, PlanID: planId}
		expectedError = nil
		asyncAllowed = true
		instanceExists = true
		instanceExistsError = nil
	})

	BeforeEach(func() {
		boshClient.DeleteDeploymentReturns(deleteTaskID, nil)
	})

	JustBeforeEach(func() {
		instanceRepo.InstanceExistsReturns(instanceExists, instanceExistsError)
		instanceRepo.FindByIDReturns(expectedInstance, expectedError)
		isAsync, deprovisionErr = b.Deprovision(instanceID, brokerapi.DeprovisionDetails{}, asyncAllowed)
	})

	It("returns that is deprovisioning asynchronously", func() {
		Expect(isAsync).To(Equal(brokerapi.IsAsync(true)))
	})

	It("deletes a bosh deployment whose name is based on the instance ID", func() {
		Expect(boshClient.DeleteDeploymentCallCount()).To(Equal(1))
		Expect(boshClient.DeleteDeploymentArgsForCall(0)).To(Equal(deploymentName(instanceID)))
	})

	It("logs that it will delete the deployment", func() {
		Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("deleting deployment for instance %s", instanceID)))
	})
	It("logs the task id to delete the deployment", func() {
		Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Delete instance %s was %d", instanceID, deleteTaskID)))
	})

	Context("when the bosh client successfully begins deleting the deployment and returns a task", func() {
		It("sets the last operation type in the instance repository", func() {
			Expect(instanceRepo.SaveCallCount()).To(Equal(1))
			Expect(instanceRepo.SaveArgsForCall(0)).To(Equal(broker.Instance{
				ID:                instanceID,
				LastOperationType: broker.Delete,
				PlanID:            planId}))
		})

		Context("when the last operation type can be set", func() {
			BeforeEach(func() {
				instanceRepo.SaveReturns(nil)
			})

			It("does not return error", func() {
				Expect(deprovisionErr).NotTo(HaveOccurred())
			})
		})

		Context("when the last operation type cannot be set", func() {
			BeforeEach(func() {
				instanceRepo.SaveReturns(errors.New("sql error"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("sql error"))
			})

			It("returns a generic error message", func() {
				Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
			})
		})
	})

	Context("when the instance does not exist", func() {
		BeforeEach(func() {
			instanceExists = false
		})

		It("returns an error", func() {
			Expect(deprovisionErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
		})

		It("doesn't delete the deployment", func() {
			Expect(boshClient.DeleteDeploymentCallCount()).To(BeZero())
		})
	})

	Context("when the checking the instance exists fails", func() {
		BeforeEach(func() {
			instanceExistsError = fmt.Errorf("io error!")
		})

		It("doesn't delete the deployment", func() {
			Expect(boshClient.DeleteDeploymentCallCount()).To(BeZero())
		})

		It("returns a generic error message", func() {
			Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("io error!"))
		})
	})

	Context("if the instance find fails", func() {
		BeforeEach(func() {
			expectedError = fmt.Errorf("problem!")
		})

		It("doesn't delete the deployment", func() {
			Expect(boshClient.DeleteDeploymentCallCount()).To(BeZero())
		})

		It("returns a generic error message", func() {
			Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("problem!"))
		})
	})

	Context("when the async allowed flag is false", func() {
		BeforeEach(func() {
			asyncAllowed = false
		})

		It("returns a  error", func() {
			Expect(deprovisionErr).To(Equal(brokerapi.ErrAsyncRequired))
		})
	})

	Context("when the bosh client cannot delete the deployment", func() {
		BeforeEach(func() {
			boshClient.DeleteDeploymentReturns(0, errors.New("er ma gerd!"))
		})

		It("returns a generic error message", func() {
			Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("er ma gerd!"))
		})
	})
})
