package broker

import (
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

const (
	CreateInProgress = "Instance provisioning in progress"
	CreateCompleted  = "Instance provisioning completed"
	DeleteInProgress = "Instance deletion in progress"
	DeleteCompleted  = "Instance deletion completed"
	UpdateCompleted  = "Instance update completed"
	UpdateInProgress = "Instance update in progress"
)

var (
	taskStates = map[string]brokerapi.LastOperationState{
		boshclient.Queued:     brokerapi.InProgress,
		boshclient.Processing: brokerapi.InProgress,
		boshclient.Done:       brokerapi.Succeeded,
		boshclient.Error:      brokerapi.Failed,
	}

	descriptions = map[brokerapi.LastOperationState]map[OperationType]string{
		brokerapi.InProgress: {
			Create:  CreateInProgress,
			Update:  UpdateInProgress,
			Upgrade: "",
			Delete:  DeleteInProgress,
		},
		brokerapi.Succeeded: {
			Create:  CreateCompleted,
			Update:  UpdateCompleted,
			Upgrade: "",
			Delete:  DeleteCompleted,
		},
		brokerapi.Failed: {
			Create:  "Instance provisioning failed: %s",
			Update:  "Instance update failed: %s",
			Upgrade: "Failed for bosh task: %d",
			Delete:  "Instance deletion failed: %s",
		},
	}
)

func constructLastOperation(boshTask boshclient.BoshTask, instance Instance) brokerapi.LastOperation {
	taskState := taskStates[boshTask.State]
	template := descriptions[taskState][instance.LastOperationType]
	if boshTask.State == boshclient.Error {

		if instance.LastOperationType == Upgrade {
			template = fmt.Sprintf(template, boshTask.ID)
		} else {
			template = fmt.Sprintf(template, genericErrorMessage)
		}
	}
	return brokerapi.LastOperation{State: taskState, Description: template}
}

func (b *Broker) LastOperation(instanceID string) (brokerapi.LastOperation, error) {
	errs := func(err error) (brokerapi.LastOperation, error) {
		return brokerapi.LastOperation{}, convertToDisplayableError(err, b.logger)
	}

	instance, err := b.instanceRepo.FindByID(instanceID)
	if err != nil {
		return errs(fmt.Errorf("error retrieving instance %s: %s", instanceID, err))
	}

	lastBoshTask, err := lastOperation(b.boshClient, deploymentName(instanceID))
	if err != nil {
		return errs(fmt.Errorf("error retrieving tasks from bosh, for deployment '%s': %s", deploymentName(instanceID), err))
	}
	lastOperation := constructLastOperation(lastBoshTask, instance)
	b.logLastOperation(instance, lastBoshTask)

	if lastBoshTask.State == boshclient.Done && instance.LastOperationType == Delete {
		if err := b.instanceRepo.DeleteByID(instanceID); err != nil {
			return errs(fmt.Errorf("error deleting instance %s: %s", instanceID, err))
		}
	}
	return lastOperation, nil
}

func (b *Broker) logLastOperation(instance Instance, boshTask boshclient.BoshTask) {
	var state string
	switch boshTask.State {
	case boshclient.Done:
		state = "success"
	case boshclient.Error:
		state = "failure"
	default:
		return
	}

	var op string
	switch instance.LastOperationType {
	case Create:
		op = "creating"
	case Update:
		op = "updating"
	case Upgrade:
		op = "upgrading"
	case Delete:
		op = "deleting"
	}

	b.logger.Printf("BOSH task ID %d status: %s %s deployment for instance %s: Description: %s Result: %s\n", boshTask.ID, state, op, instance.ID, boshTask.Description, boshTask.Result)
}

func lastOperation(boshClient BoshClient, deploymentName string) (boshclient.BoshTask, error) {
	boshTasksForDeployment, err := boshClient.GetTasks(deploymentName)
	if err != nil {
		return boshclient.BoshTask{}, err
	}

	if len(boshTasksForDeployment) == 0 {
		return boshclient.BoshTask{}, noTasksError
	}

	return boshTasksForDeployment[0], nil
}

var noTasksError = fmt.Errorf("no tasks for deployment")
