package broker_test

import (
	"errors"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("LastOperation", func() {
	Context("failures", func() {
		var (
			instanceID = "a-useful-instance"

			lastOpErr error
		)

		BeforeEach(func() {
			instanceRepo.FindByIDReturns(broker.Instance{ID: instanceID}, nil)
		})

		JustBeforeEach(func() {
			_, lastOpErr = b.LastOperation(instanceID)
		})

		Context("when no task exists for that instance ID", func() {
			BeforeEach(func() {
				instanceRepo.FindByIDReturns(broker.Instance{}, errors.New("no task exists!"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error retrieving instance %s: no task exists!", instanceID)))
			})
			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})
		})

		Context("when there is a error retrieving the instance", func() {
			BeforeEach(func() {
				instanceRepo.FindByIDReturns(broker.Instance{}, errors.New("some error"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error retrieving instance %s: some error", instanceID)))
			})
			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when tasks cannot be retrieved from BOSH", func() {
			BeforeEach(func() {
				boshClient.GetTasksReturns(nil, errors.New("something went wrong!"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error retrieving tasks from bosh, for deployment 'service-instance_a-useful-instance': something went wrong!"))
			})
			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when not tasks found on BOSH", func() {
			BeforeEach(func() {
				boshClient.GetTasksReturns([]boshclient.BoshTask{}, nil)
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error retrieving tasks from bosh, for deployment 'service-instance_a-useful-instance': no tasks for deployment"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})
	})

	Context("when the last task can be retrieved", func() {
		var instanceID = "not-relevent"
		var taskID = 199
		type testCase struct {
			ActualBoshTask      boshclient.BoshTask
			ActualOperationType broker.OperationType
			LogContains         string

			ExpectedLastOperation       brokerapi.LastOperation
			TaskIDRemoved               bool
			InstanceDeletedSuccessfully bool
		}

		testLastOperation := func(testCase testCase) func() {
			return func() {
				var (
					actualLastOperation      brokerapi.LastOperation
					actualLastOperationError error
				)

				JustBeforeEach(func() {
					instanceRepo.FindByIDReturns(broker.Instance{ID: instanceID, LastOperationType: testCase.ActualOperationType}, nil)
					boshClient.GetTasksReturns([]boshclient.BoshTask{testCase.ActualBoshTask}, nil)

					actualLastOperation, actualLastOperationError = b.LastOperation(instanceID)
				})

				It("retrieves the task ID for the instance", func() {
					Expect(instanceRepo.FindByIDCallCount()).To(Equal(1))
					Expect(instanceRepo.FindByIDArgsForCall(0)).To(Equal(instanceID))
				})

				It("does not error", func() {
					Expect(actualLastOperationError).NotTo(HaveOccurred())
				})

				It("returns the state", func() {
					Expect(actualLastOperation.State).To(Equal(testCase.ExpectedLastOperation.State))
				})

				It("returns the description", func() {
					Expect(actualLastOperation.Description).To(Equal(testCase.ExpectedLastOperation.Description))
				})

				It("does not save the instance", func() {
					Expect(instanceRepo.SaveCallCount()).To(Equal(0))
				})

				It("logs a message", func() {
					Expect(logBuffer.String()).To(ContainSubstring(testCase.LogContains))
				})

				if testCase.InstanceDeletedSuccessfully {
					It("deletes the instance from the instance repository", func() {
						Expect(instanceRepo.DeleteByIDCallCount()).To(Equal(1))
						Expect(instanceRepo.DeleteByIDArgsForCall(0)).To(Equal(instanceID))
					})
				} else {
					It("does not delete the instance from the instance repository", func() {
						Expect(instanceRepo.DeleteByIDCallCount()).To(Equal(0))
					})
				}

				if testCase.ActualOperationType == broker.Create {
					if testCase.ActualBoshTask.State == boshclient.Done {
						It("logs that the deployment has been created", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success creating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.Error {
						It("logs that the deployment has been failed to create", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure creating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				} else if testCase.ActualOperationType == broker.Delete {
					if testCase.ActualBoshTask.State == boshclient.Done {
						It("logs that the deployment has been deleted", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success deleting deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.Error {
						It("logs that the deployment has been failed to delete", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure deleting deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				} else if testCase.ActualOperationType == broker.Update {
					if testCase.ActualBoshTask.State == boshclient.Done {
						It("logs that the deployment has been updated", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success updating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.Error {
						It("logs that the deployment has been failed to update", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure updating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				} else if testCase.ActualOperationType == broker.Upgrade {
					if testCase.ActualBoshTask.State == boshclient.Done {
						It("logs that the deployment has been upgraded", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success upgrading deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.Error {
						It("logs that the deployment has been failed to upgrade", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure upgrading deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				}
			}
		}
		Describe("while creating ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Processing, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Create,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.CreateInProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Queued, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Create,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.CreateInProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Error, Result: "not good", Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Create,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Done, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Create,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: broker.CreateCompleted,
					},
					TaskIDRemoved: true,
				}))
		})

		Describe("while deleting ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Processing, Description: boshclient.BoshTaskDeleteDeploymentPrefix + "-" + instanceID},
					ActualOperationType: broker.Delete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.DeleteInProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Queued, Description: boshclient.BoshTaskDeleteDeploymentPrefix + "-" + instanceID},
					ActualOperationType: broker.Delete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.DeleteInProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Error, Result: "not good", Description: boshclient.BoshTaskDeleteDeploymentPrefix + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.Delete,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Done, Description: boshclient.BoshTaskDeleteDeploymentPrefix + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.Delete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: broker.DeleteCompleted,
					},
					InstanceDeletedSuccessfully: true,
				}))
		})

		Describe("while updating ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Processing, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Update,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.UpdateInProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Queued, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Update,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.UpdateInProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Error, Result: "not good", Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Update,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Done, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Update,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: broker.UpdateCompleted,
					},
					TaskIDRemoved: true,
				}))
		})

		Describe("while upgrading ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Processing, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Upgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State: brokerapi.InProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Queued, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Upgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State: brokerapi.InProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Error, Result: "not good", Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Upgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.Done, Description: boshclient.BoshTaskCreateDeployment, ID: taskID},
					ActualOperationType: broker.Upgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State: brokerapi.Succeeded,
					},
					TaskIDRemoved: true,
				}))
		})
	})

})
