package broker

import (
	"fmt"

	"github.com/pivotal-cf/brokerapi"
)

func (b *Broker) Provision(instanceID string, details brokerapi.ProvisionDetails, asyncAllowed bool) (brokerapi.ProvisionedServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	if !asyncAllowed {
		return brokerapi.ProvisionedServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	requestParams, err := convertProvisionDetailsToMap(details)
	if err != nil {
		return brokerapi.ProvisionedServiceSpec{}, err
	}

	dashboardURL, err := b.deployInstance(instanceID, details.PlanID, "", requestParams)

	return brokerapi.ProvisionedServiceSpec{IsAsync: true, DashboardURL: dashboardURL}, convertToDisplayableError(err, b.logger)
}

func (b *Broker) deployInstance(instanceID, planID, previousPlanID string, arbitraryParams map[string]interface{}) (string, error) {
	errs := func(err error) (string, error) {
		b.logger.Println(err)
		return "", err
	}

	planConfig := b.serviceOffering.FindPlanById(planID)
	if planConfig == nil {
		return errs(DisplayableError{fmt.Errorf("plan %s not found", planID)})
	}

	exists, err := b.instanceRepo.InstanceExists(instanceID)
	if err != nil {
		return errs(fmt.Errorf("error fetching instance: %s", err))
	}
	if exists {
		return errs(DisplayableError{brokerapi.ErrInstanceAlreadyExists})
	}

	count, err := b.instanceRepo.CountByPlanID(planConfig.ID)
	if err != nil {
		return errs(fmt.Errorf("error counting plans: %s", err))
	}

	if planConfig.Quotas != nil && count >= planConfig.Quotas.ServiceInstanceLimit {
		return errs(DisplayableError{brokerapi.ErrPlanQuotaExceeded})
	}

	_, manifest, err := b.deployer.Deploy(Instance{ID: instanceID, PlanID: planConfig.ID}, planID, deploymentName(instanceID), Create, arbitraryParams, nil, nil)
	if err != nil {
		return errs(err)
	}

	dashboardUrl, err := b.serviceAdapter.GenerateDashboardUrl(instanceID, serviceAdapterPlan(*planConfig), manifest)
	switch err := err.(type) {
	case nil, ServiceAdapterCommandNotImplementedError:
		return dashboardUrl, nil
	case error:
		return "", err
	}

	return dashboardUrl, nil
}
