package broker_test

import (
	"errors"

	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

type ItRedeploysContext struct {
	instanceID            string
	expectedRequestParams map[string]interface{}
	oldPlanID             string
	newPlanID             string
	redeployErr           error
	operationType         broker.OperationType
}

func ItRedeploys(context *ItRedeploysContext) {
	var (
		instanceID            = context.instanceID
		serviceDeploymentName = deploymentName(instanceID)
		requestParams         = context.expectedRequestParams

		expectedPreviousManifest = []byte("old-manifest-fetched-from-bosh")
		oldPlanID                = context.oldPlanID
		taskID                   = 42
		newPlanID                = context.newPlanID
		existingInstanceInDB     broker.Instance
		operationType            = context.operationType
	)

	BeforeEach(func() {
		existingInstanceInDB = broker.Instance{ID: instanceID, PlanID: oldPlanID}
		instanceRepo.FindByIDReturns(existingInstanceInDB, nil)
	})

	Context("deployment goes well", func() {
		var manifest = []byte("updated-manifest")

		BeforeEach(func() {
			boshClient.DeployRawReturns(taskID, nil)
			instanceRepo.SaveReturns(nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetManifestRawReturns(expectedPreviousManifest, nil)

			fakeDeployer.DeployReturns(broker.Instance{}, []byte("new-manifest-fetched-from-adapter"), nil)
		})

		It("does not error", func() {
			Expect(context.redeployErr).NotTo(HaveOccurred())
		})

		It("fetches correct instance", func() {
			// fetches correct instance
			Expect(instanceRepo.FindByIDCallCount()).To(Equal(1))
			Expect(instanceRepo.FindByIDArgsForCall(0)).To(Equal(instanceID))
		})

		It("fetches correct manifest from bosh", func() {
			Expect(boshClient.GetManifestRawCallCount()).To(Equal(1))
			Expect(boshClient.GetManifestRawArgsForCall(0)).To(Equal(serviceDeploymentName))
		})

		It("deploys with the new planID", func() {
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
			actualInstance, actualPlanID, actualDeploymentName, actualOperationType, actualReqeustParams, actualOldManifest, actualPreviousPlanID := fakeDeployer.DeployArgsForCall(0)
			Expect(actualInstance).To(Equal(existingInstanceInDB))
			Expect(actualPlanID).To(Equal(newPlanID))
			Expect(actualDeploymentName).To(Equal(broker.InstancePrefix + instanceID))
			Expect(actualReqeustParams).To(Equal(requestParams))
			Expect(actualOldManifest).To(Equal(expectedPreviousManifest))
			Expect(actualOperationType).To(Equal(operationType))
			oldPlanIDCopy := oldPlanID
			Expect(actualPreviousPlanID).To(Equal(&oldPlanIDCopy))
		})

		Context("when the plan quota has been exceeded", func() {
			BeforeEach(func() {
				instanceRepo.CountByPlanIDReturns(serviceInstanceLimit+1, nil)
			})

			It("updates anyway", func() {
				Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
			})
		})
	})

	Context("when bosh fails to fetch previous manifest", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, errors.New("no manifest for you"))
		})
		if context.operationType == broker.Update {
			It("returns a generic error message to the cli user", func() {
				Expect(context.redeployErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("no manifest for you"))
			})
		}
		if context.operationType == broker.Upgrade {
			It("returns a generic error message to the cli user", func() {
				Expect(context.redeployErr).To(MatchError(ContainSubstring("no manifest for you")))
			})
		}
	})

	Context("when find instance by id fails", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(expectedPreviousManifest, nil)
			instanceRepo.FindByIDReturns(broker.Instance{}, errors.New("no instance for you"))
		})
		if context.operationType == broker.Update {
			It("returns a generic error message to the cli user", func() {
				Expect(context.redeployErr).To(MatchError(ContainSubstring("An internal error has occurred")))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("no instance for you"))
			})
		}

		if context.operationType == broker.Upgrade {
			It("returns a generic error message to the cli user", func() {
				Expect(context.redeployErr).To(MatchError(ContainSubstring("no instance for you")))
			})
		}
	})

}
