package broker_test

import (
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("Unbind", func() {
	var (
		instanceID     = "a-most-unimpressive-instance"
		bindingID      = "I'm still a binding"
		deploymentName = broker.InstancePrefix + instanceID
		boshVms        = bosh.BoshVMs{"redis-server": []string{"an.ip"}}
		actualManifest = []byte("a valid manifest")
		unbindErr      error
	)

	BeforeEach(func() {
		boshClient.VMsReturns(boshVms, nil)
		serviceAdapter.DeleteBindingReturns(nil)
		instanceRepo.InstanceExistsReturns(true, nil)
		boshClient.GetManifestRawReturns(actualManifest, nil)
	})

	JustBeforeEach(func() {
		unbindErr = b.Unbind(instanceID, bindingID, brokerapi.UnbindDetails{})
	})

	It("asks bosh for VMs from a deployment named by the manifest generator", func() {
		Expect(boshClient.VMsCallCount()).To(Equal(1))
		Expect(boshClient.VMsArgsForCall(0)).To(Equal(deploymentName))
	})

	It("destroys the binding using the bosh topology and admin credentials", func() {
		Expect(serviceAdapter.DeleteBindingCallCount()).To(Equal(1))
		passedBindingID, passedVms, passedManifest := serviceAdapter.DeleteBindingArgsForCall(0)
		Expect(passedBindingID).To(Equal(bindingID))
		Expect(passedVms).To(Equal(boshVms))
		Expect(passedManifest).To(Equal(actualManifest))
	})

	It("does not error", func() {
		Expect(unbindErr).NotTo(HaveOccurred())
	})

	Context("when bosh fails to get VMs", func() {
		BeforeEach(func() {
			boshClient.VMsReturns(nil, errors.New("oops"))
		})

		It("returns a generic error message to the cli user", func() {
			Expect(unbindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("oops"))
		})
	})

	Context("when bosh client does not return a manifest", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, nil)
		})

		It("returns an error", func() {
			Expect(unbindErr).To(HaveOccurred())
		})
	})

	Context("when there is an error while fetching a manifest from the bosh client", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns(nil, fmt.Errorf("problem fetching manifest"))
		})

		It("returns an error", func() {
			Expect(unbindErr).To(HaveOccurred())
		})
	})

	Context("when cannot find the instance", func() {
		BeforeEach(func() {
			instanceRepo.InstanceExistsReturns(false, nil)
		})

		It("returns an error", func() {
			Expect(unbindErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
		})
	})

	Context("when the service adapter fails to destroy the binding", func() {
		BeforeEach(func() {
			serviceAdapter.DeleteBindingReturns(errors.New("unbinding fail"))
		})

		It("returns a generic error message to the cli user", func() {
			Expect(unbindErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("unbinding fail"))
		})
	})
})
