package main

import (
	"encoding/json"
	"flag"
	"fmt"
	"log"
	"os/exec"
	"time"
)

func main() {
	serviceUniqueID := flag.String("serviceID", "", "Service ID")
	timeoutMinutes := flag.Int("timeoutMinutes", 0, "time limit in minutes")
	flag.Parse()

	if *serviceUniqueID == "" {
		panic("-serviceID flag must be set")
	}
	if *timeoutMinutes == 0 {
		panic("-timeoutMinutes flag must be set")
	}

	instanceURLs := []string{}

	planGUIDs := getPlanGUIDs(*serviceUniqueID)
	for _, plan := range planGUIDs {
		instancesOutput := cfCurl(fmt.Sprintf("/v2/service_instances?q=service_plan_guid:%s", plan), "GET")
		var serviceInstancesResp ServiceInstancesResponse
		must(json.Unmarshal(instancesOutput, &serviceInstancesResp))

		for _, instance := range serviceInstancesResp.Resources {
			deleteBindings(instance.Entity.ServiceBindingsURL)
			deleteServiceKeys(instance.Entity.ServiceKeysURL)

			log.Printf("Deleting service instance %s (%s)\n", instance.Metadata.GUID, instance.Entity.Name)
			cfCurl(fmt.Sprintf("%s?accepts_incomplete=true", instance.Metadata.URL), "DELETE")

			instanceURLs = append(instanceURLs, instance.Metadata.URL)
		}
	}

	if len(instanceURLs) == 0 {
		return
	}

	select {
	case <-blockUntilAllInstancesGone(instanceURLs):
	case <-time.After(time.Minute * time.Duration(*timeoutMinutes)):
		panic("not all instances deleted after 1 hour")
	}

	errorIfAnyInstancesLeft(planGUIDs)
}

func getPlanGUIDs(serviceID string) []string {
	var services GuidAndUniqueIDResponse
	must(json.Unmarshal(cfCurl("/v2/services", "GET"), &services))

	serviceGUID := findForUUID(services, serviceID)

	plansOutput := cfCurl(fmt.Sprintf("/v2/services/%s/service_plans", serviceGUID), "GET")
	var plansResp GuidAndUniqueIDResponse
	must(json.Unmarshal(plansOutput, &plansResp))

	planGUIDs := []string{}
	for _, plan := range plansResp.Resources {
		planGUIDs = append(planGUIDs, plan.Metadata.GUID)
	}

	return planGUIDs
}

func findForUUID(resources GuidAndUniqueIDResponse, uuid string) string {
	for _, service := range resources.Resources {
		if service.Entity.UniqueID == uuid {
			return service.Metadata.GUID
		}
	}

	panic(fmt.Errorf("no resource found with UUID %s", uuid))
}

func deleteBindings(instanceBindingsURL string) {
	bindingsOutput := cfCurl(instanceBindingsURL, "GET")
	var bindingsResp BindingsResponse
	must(json.Unmarshal(bindingsOutput, &bindingsResp))
	for _, binding := range bindingsResp.Resources {
		log.Printf("Deleting binding %s of service instance %s to app %s\n", binding.Metadata.GUID, binding.Entity.ServiceInstanceGUID, binding.Entity.AppGUID)
		cfCurl(fmt.Sprintf("/v2/apps/%s/service_bindings/%s", binding.Entity.AppGUID, binding.Metadata.GUID), "DELETE")
	}
}

func deleteServiceKeys(serviceKeysURL string) {
	serviceKeysOutput := cfCurl(serviceKeysURL, "GET")
	var serviceKeysResp ServiceKeysResponse
	must(json.Unmarshal(serviceKeysOutput, &serviceKeysResp))

	for _, serviceKey := range serviceKeysResp.Resources {
		log.Printf("Deleting service key %s of service instance %s\n", serviceKey.Metadata.GUID, serviceKey.Entity.ServiceInstanceGUID)
		cfCurl(serviceKey.Metadata.URL, "DELETE")
	}
}

func blockUntilAllInstancesGone(instanceURLs []string) chan bool {
	done := make(chan bool)
	go func() {
		instanceCount := len(instanceURLs)

		i := -1
		successiveDeadInstances := 0
		for {
			i = (i + 1) % instanceCount
			if successiveDeadInstances >= instanceCount {
				break
			}

			getInstanceOutput := cfCurlAllowErrors(instanceURLs[i], "GET")

			possibleError := make(map[string]interface{})
			must(json.Unmarshal(getInstanceOutput, &possibleError))

			if _, ok := possibleError["error_code"]; ok {
				successiveDeadInstances++
				continue
			}

			var instanceResp ServiceInstanceResource
			must(json.Unmarshal(getInstanceOutput, &instanceResp))

			if instanceResp.Entity.LastOperation.State == "failed" {
				panic("failed to delete service instance")
			}

			successiveDeadInstances = 0
			time.Sleep(time.Second * 5)
		}
		done <- true
	}()
	return done
}

func errorIfAnyInstancesLeft(planGUIDs []string) {
	for _, plan := range planGUIDs {
		instancesOutput := cfCurl(fmt.Sprintf("/v2/service_instances?q=service_plan_guid:%s", plan), "GET")
		var serviceInstancesResp ServiceInstancesResponse
		must(json.Unmarshal(instancesOutput, &serviceInstancesResp))
		if len(serviceInstancesResp.Resources) != 0 {
			panic(fmt.Errorf("expected 0 instances for plan GUID %s, got %d", plan, len(serviceInstancesResp.Resources)))
		}
	}
}

func errorResponse(resp []byte) error {
	if len(resp) < 2 {
		// if empty response, then it is not an error
		return nil
	}

	possibleError := make(map[string]interface{})
	if err := json.Unmarshal(resp, &possibleError); err != nil {
		return err
	}

	if errorCode, ok := possibleError["error_code"]; ok {
		return fmt.Errorf("code: %s, description %s", errorCode.(string), possibleError["description"].(string))
	}

	return nil
}

func cfCurl(path, method string) []byte {
	out, err := exec.Command("cf", "curl", path, "-X", method).CombinedOutput()
	if err != nil {
		panic(fmt.Errorf("an error occurred executing a CF API command: %s, output: %s", err, string(out)))
	}
	mustNot(errorResponse(out))
	return out
}

func cfCurlAllowErrors(path, method string) []byte {
	out, err := exec.Command("cf", "curl", path, "-X", method).CombinedOutput()
	mustNot(err)
	return out
}

func must(err error) {
	if err != nil {
		panic(err)
	}
}

var mustNot = must

type ServiceInstancesResponse struct {
	Resources []ServiceInstanceResource
}

type ServiceInstanceResource struct {
	Metadata Metadata
	Entity   ServiceInstanceEntity
}

type Metadata struct {
	GUID string
	URL  string
}

type ServiceInstanceEntity struct {
	Name               string
	ServiceBindingsURL string        `json:"service_bindings_url"`
	ServiceKeysURL     string        `json:"service_keys_url"`
	LastOperation      LastOperation `json:"last_operation"`
}

type LastOperation struct {
	State string
}

type GuidAndUniqueIDResponse struct {
	Resources []struct {
		Metadata struct {
			GUID string
		}
		Entity struct {
			UniqueID string `json:"unique_id"`
		}
	}
}

type BindingsResponse struct {
	Resources []BindingResource
}

type BindingResource struct {
	Metadata struct {
		GUID string
	}
	Entity struct {
		AppGUID             string `json:"app_guid"`
		ServiceInstanceGUID string `json:"service_instance_guid"`
	}
}

type ServiceKeysResponse struct {
	Resources []struct {
		Metadata struct {
			GUID string
			URL  string
		}
		Entity struct {
			ServiceInstanceGUID string `json:"service_instance_guid"`
		}
	}
}
