package main

import (
	"database/sql"
	"flag"
	"fmt"
	"log"
	"net/http"
	"os"
	"os/signal"
	"syscall"

	"github.com/gorilla/mux"
	apiauth "github.com/pivotal-cf/brokerapi/auth"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient/auth"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/config"
	"github.com/pivotal-cf/on-demand-service-broker/external"
	"github.com/pivotal-cf/on-demand-service-broker/instance_repository"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"

	"github.com/codegangsta/negroni"
	_ "github.com/mattn/go-sqlite3"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-golang/lager"
)

func main() {
	// Yes, you really do need the space at the end of the prefix.
	logger := log.New(os.Stdout, "[on-demand-service-broker] ", log.LstdFlags)
	logger.Println("Starting broker")

	configFilePath := flag.String("configFilePath", "", "path to config file")

	flag.Parse()
	if *configFilePath == "" {
		logger.Fatal("must supply -configFilePath")
	}

	conf, err := config.Parse(*configFilePath)
	if err != nil {
		logger.Fatalf("error parsing config: %s", err)
	}

	var (
		authenticator boshclient.AuthHeaderBuilder
	)

	authConfig := conf.Bosh.Authentication
	if authConfig.Basic.IsSet() {
		authenticator = auth.NewBasicAuthHeaderBuilder(
			authConfig.Basic.Username,
			authConfig.Basic.Password,
		)
	} else if authConfig.UAA.IsSet() {
		authenticator = auth.NewBearerAuthHeaderBuilder(
			authConfig.UAA.UAAURL,
			authConfig.UAA.ID,
			authConfig.UAA.Secret,
			conf.Broker.DisableSSLCertVerification,
			[]byte(conf.Bosh.TrustedCert),
		)
	}

	boshClient := boshclient.New(conf.Bosh.URL, authenticator, logger, conf.Broker.DisableSSLCertVerification, []byte(conf.Bosh.TrustedCert))

	dbConn, err := sql.Open("sqlite3", conf.Broker.SQLiteDBPath)
	if err != nil {
		logger.Fatalf("error opening sqlite database: %s", err)
	}

	onExit := make(chan os.Signal, 1)
	signal.Notify(onExit, os.Interrupt)
	signal.Notify(onExit, syscall.SIGTERM)
	go func() {
		<-onExit
		logger.Println("Stopping broker")
		if err := dbConn.Close(); err != nil {
			logger.Printf("error closing database connection: %s\n", err)
		}
		os.Exit(1)
	}()

	instanceRepository := instance_repository.New(dbConn)

	serviceAdapter := &external.Adapter{
		ExternalBinPath: conf.ServiceAdapter.Path,
		CommandRunner:   external.NewCommandRunner(),
		Logger:          logger,
	}

	deploymentManager := broker.NewDeployer(
		boshClient,
		instanceRepository,
		serviceAdapter,
		conf.ServiceCatalog,
		conf.BaseDeploymentProperties.Stemcell,
		conf.BaseDeploymentProperties.Releases,
		logger,
	)
	onDemandBroker, err := broker.New(boshClient, instanceRepository, serviceAdapter, deploymentManager, conf.ServiceCatalog, logger)
	if err != nil {
		logger.Fatalf("error starting broker: %s", err)
	}

	brokerRouter := mux.NewRouter()

	mgmtapi.AttachRoutes(brokerRouter, onDemandBroker, logger)
	brokerapi.AttachRoutes(brokerRouter, onDemandBroker, lager.NewLogger("on-demand-service-broker"))
	authProtectedBrokerAPI := apiauth.NewWrapper(conf.Broker.Username, conf.Broker.Password).Wrap(brokerRouter)

	// can't use negroni.Classic() if we want custom logger flags in Negroni
	negroniLogger := negroni.NewLogger()
	negroniLogger.Logger = log.New(os.Stdout, "[negroni] ", log.LstdFlags)
	server := negroni.New(negroni.NewRecovery(), negroniLogger, negroni.NewStatic(http.Dir("public")))

	server.UseHandler(authProtectedBrokerAPI)
	server.Run(fmt.Sprintf("0.0.0.0:%d", conf.Broker.Port))
}
