package config

import (
	"errors"
	"fmt"
	"io/ioutil"
	"reflect"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/broker"

	"gopkg.in/yaml.v2"
)

type Config struct {
	Broker                   Broker
	Bosh                     Bosh
	ServiceAdapter           ServiceAdapter           `yaml:"service_adapter"`
	BaseDeploymentProperties BaseDeploymentProperties `yaml:"service_deployment"`
	ServiceCatalog           broker.ServiceCatalog    `yaml:"service_catalog"`
}

func (c Config) Validate() error {
	if err := c.Broker.Validate(); err != nil {
		return err
	}

	if err := c.Bosh.Validate(); err != nil {
		return err
	}

	if err := checkIsExecutableFile(c.ServiceAdapter.Path); err != nil {
		return err
	}

	return nil
}

type Broker struct {
	Port                       int
	Username                   string
	Password                   string
	SQLiteDBPath               string `yaml:"sqlite_db_path"`
	DisableSSLCertVerification bool   `yaml:"disable_ssl_cert_verification"`
}

func (b Broker) Validate() error {
	if b.Port == 0 {
		return errors.New("broker.port can't be empty")
	}
	if b.Username == "" {
		return errors.New("broker.username can't be empty")
	}
	if b.Password == "" {
		return errors.New("broker.password can't be empty")
	}

	return nil
}

type BaseDeploymentProperties struct {
	Releases serviceadapter.ServiceReleases
	Stemcell serviceadapter.Stemcell
}

type Bosh struct {
	URL            string
	TrustedCert    string `yaml:"root_ca_cert"`
	Authentication BoshAuthentication
}

type BoshAuthentication struct {
	Basic Credentials
	UAA   ClientCredentials
}

type Credentials struct {
	Username string
	Password string
}

func (c Credentials) IsSet() bool {
	return c != Credentials{}
}

type ClientCredentials struct {
	UAAURL string `yaml:"url"`
	ID     string `yaml:"client_id"`
	Secret string `yaml:"client_secret"`
}

func (cc ClientCredentials) IsSet() bool {
	return cc != ClientCredentials{}
}

func (b Bosh) Validate() error {
	return b.Authentication.Validate()
}

func (a BoshAuthentication) Validate() error {
	uaaSet := a.UAA.IsSet()
	basicSet := a.Basic.IsSet()

	var err error

	if !uaaSet && !basicSet {
		return fmt.Errorf("Must specify bosh authentication")
	} else if uaaSet && basicSet {
		err = fmt.Errorf("Cannot specify both basic auth and UAA authentication")
	} else if uaaSet {
		err = validateNoFieldsEmptyString(a.UAA, "UAA")
	} else if basicSet {
		err = validateNoFieldsEmptyString(a.Basic, "Basic")
	}

	return err
}

type ServiceAdapter struct {
	Path string
}

func Parse(configFilePath string) (Config, error) {
	configFileBytes, err := ioutil.ReadFile(configFilePath)
	if err != nil {
		return Config{}, err
	}

	var config Config
	if err := yaml.Unmarshal(configFileBytes, &config); err != nil {
		return Config{}, err
	}

	if err := config.Validate(); err != nil {
		return Config{}, err
	}

	return config, nil
}

func validateNoFieldsEmptyString(obj interface{}, objName string) error {
	bVal := reflect.ValueOf(obj)
	bType := reflect.TypeOf(obj)
	for i := 0; i < bVal.NumField(); i++ {
		fieldVal := bVal.Field(i).String()

		if fieldVal == "" {
			fieldName := bType.Field(i).Name
			return fmt.Errorf("%s.%s can't be empty", objName, fieldName)
		}
	}

	return nil
}
