package config_test

import (
	"os"
	"path/filepath"

	"gopkg.in/yaml.v2"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/config"
)

var _ = Describe("BrokerConfig", func() {
	var (
		configFileName string
		conf           config.Config
		parseErr       error
	)

	JustBeforeEach(func() {
		cwd, err := os.Getwd()
		Expect(err).ToNot(HaveOccurred())
		configFilePath := filepath.Join(cwd, "test_assets", configFileName)
		conf, parseErr = config.Parse(configFilePath)
	})

	Context("when a valid yaml config file is supplied", func() {
		Context("when the BOSH director uses basic auth", func() {
			BeforeEach(func() {
				configFileName = "good_config.yml"
			})

			It("returns config object", func() {
				expected := config.Config{
					Broker: config.Broker{
						Port:                       8080,
						Username:                   "username",
						Password:                   "password",
						DisableSSLCertVerification: true,
					},
					Bosh: config.Bosh{
						URL:         "some-url",
						TrustedCert: "some-cert",
						Authentication: config.BoshAuthentication{
							Basic: config.Credentials{
								Username: "some-username",
								Password: "some-password",
							},
						},
					},
					ServiceAdapter: config.ServiceAdapter{
						Path: "test_assets/executable.sh",
					},
					BaseDeploymentProperties: config.BaseDeploymentProperties{
						Releases: serviceadapter.ServiceReleases{{
							Name:    "some-name",
							Version: "some-version",
							Jobs:    []string{"some-job"},
						}},
						Stemcell: serviceadapter.Stemcell{OS: "Windows", Version: "XP"},
					},
					ServiceCatalog: broker.ServiceCatalog{
						ID:            "some-id",
						Name:          "some-marketplace-name",
						Description:   "some-description",
						Bindable:      true,
						PlanUpdatable: true,
						Metadata: broker.ServiceMetadata{
							DisplayName:         "some-service-display-name",
							ImageURL:            "http://test.jpg",
							LongDescription:     "Some description",
							ProviderDisplayName: "some name",
							DocumentationURL:    "some url",
							SupportURL:          "some url",
						},
						DashboardClient: &broker.DashboardClient{
							ID:          "client-id-1",
							Secret:      "secret-1",
							RedirectUri: "https://dashboard.url",
						},

						Tags: []string{"some-tag", "some-other-tag"},
						Plans: []broker.Plan{
							{
								ID:          "some-dedicated-plan-id",
								Name:        "some-dedicated-name",
								Description: "I'm a dedicated plan",
								Free:        booleanPointer(true),
								Metadata: broker.PlanMetadata{
									DisplayName: "Dedicated-Cluster",
									Bullets:     []string{"bullet one", "bullet two", "bullet three"},
									Costs: []broker.PlanCost{
										{Amount: map[string]float64{"usd": 99.0, "eur": 49.0}, Unit: "MONTHLY"},
										{Amount: map[string]float64{"usd": 0.99, "eur": 0.49}, Unit: "1GB of messages over 20GB"},
									},
								},
								Quotas: &broker.Quotas{
									ServiceInstanceLimit: 1,
								},
								Properties: serviceadapter.Properties{
									"persistence": true,
								},
								InstanceGroups: []serviceadapter.InstanceGroup{
									{
										Name:           "redis-server",
										VMType:         "some-vm",
										PersistentDisk: "some-disk",
										Instances:      34,
										Networks:       []string{"net1", "net2"},
									},
									{
										Name:      "redis-server-2",
										VMType:    "some-vm-2",
										Instances: 3,
										Networks:  []string{"net4", "net5"},
									},
									{
										Name:      "redis-errand",
										VMType:    "some-vm-3",
										Instances: 1,
										Networks:  []string{"net5", "net6"},
										Lifecycle: "errand",
									},
								},
							},
						},
					},
				}

				Expect(conf).To(Equal(expected))
			})

			It("returns no error", func() {
				Expect(parseErr).NotTo(HaveOccurred())
			})
		})

		Context("when service catalog has optional 'requires' field", func() {
			BeforeEach(func() {
				configFileName = "config_with_requires.yml"
			})

			It("returns config object", func() {
				expected := config.Config{
					Broker: config.Broker{
						Port:     8080,
						Username: "username",
						Password: "password",
					},
					Bosh: config.Bosh{
						URL: "some-url",
						Authentication: config.BoshAuthentication{
							Basic: config.Credentials{
								Username: "some-username",
								Password: "some-password",
							},
						},
					},
					ServiceAdapter: config.ServiceAdapter{
						Path: "test_assets/executable.sh",
					},
					BaseDeploymentProperties: config.BaseDeploymentProperties{
						Releases: serviceadapter.ServiceReleases{{
							Name:    "some-name",
							Version: "some-version",
							Jobs:    []string{"some-job"},
						}},
						Stemcell: serviceadapter.Stemcell{OS: "Windows", Version: "XP"},
					},
					ServiceCatalog: broker.ServiceCatalog{
						ID:            "some-id",
						Name:          "some-marketplace-name",
						Description:   "some-description",
						Bindable:      true,
						PlanUpdatable: true,
						Requires:      []string{"syslog_drain", "route_forwarding"},
						Metadata: broker.ServiceMetadata{
							DisplayName: "some-service-display-name",
						},
						Tags: []string{"some-tag", "some-other-tag"},
						Plans: []broker.Plan{
							{
								ID:          "some-dedicated-plan-id",
								Name:        "some-dedicated-name",
								Description: "I'm a dedicated plan",
								Metadata: broker.PlanMetadata{
									DisplayName: "Dedicated-Cluster",
									Bullets:     []string{"bullet one", "bullet two", "bullet three"},
								},
								Quotas: &broker.Quotas{
									ServiceInstanceLimit: 1,
								},
								Properties: serviceadapter.Properties{
									"persistence": true,
								},
								InstanceGroups: []serviceadapter.InstanceGroup{
									{
										Name:           "redis-server",
										VMType:         "some-vm",
										PersistentDisk: "some-disk",
										Instances:      34,
										Networks:       []string{"net1", "net2"},
									},
									{
										Name:      "redis-server-2",
										VMType:    "some-vm-2",
										Instances: 3,
										Networks:  []string{"net4", "net5"},
									},
								},
							},
						},
					},
				}
				Expect(conf).To(Equal(expected))
			})

			It("returns no error", func() {
				Expect(parseErr).NotTo(HaveOccurred())
			})
		})

		Context("when the BOSH director uses UAA", func() {
			BeforeEach(func() {
				configFileName = "bosh_uaa_config.yml"
			})

			It("returns no error", func() {
				Expect(parseErr).NotTo(HaveOccurred())
			})

			It("returns a config object", func() {
				Expect(conf.Bosh.Authentication.UAA).To(Equal(config.ClientCredentials{
					UAAURL: "http://some-uaa-server:99",
					ID:     "some-client-id",
					Secret: "some-client-secret",
				}))
			})
		})

		Context("when the configuration contains a non-executable service adapter", func() {
			BeforeEach(func() {
				configFileName = "config_without_executable_adapter.yml"
			})

			It("returns an error", func() {
				Expect(parseErr).To(MatchError(ContainSubstring("not executable")))
			})
		})

		Context("when the configuration does not specify any BOSH authentication", func() {
			BeforeEach(func() {
				configFileName = "bosh_no_auth_config.yml"
			})

			It("returns an error", func() {
				Expect(parseErr).To(MatchError(ContainSubstring("Must specify bosh authentication")))
			})
		})

		Context("when the configuration specified incomplete basic BOSH authentication", func() {
			BeforeEach(func() {
				configFileName = "bosh_bad_basic_auth_config.yml"
			})

			It("returns an error", func() {
				Expect(parseErr).To(MatchError(ContainSubstring("Basic.Password can't be empty")))
			})
		})

		Context("when the configuration specified incomplete UAA BOSH authentication", func() {
			BeforeEach(func() {
				configFileName = "bosh_bad_uaa_auth_config.yml"
			})

			It("returns an error", func() {
				Expect(parseErr).To(MatchError(ContainSubstring("UAA.Secret can't be empty")))
			})
		})

		Context("when the configuration specifies both types of BOSH authentication", func() {
			BeforeEach(func() {
				configFileName = "bosh_both_auth_config.yml"
			})

			It("returns an error", func() {
				Expect(parseErr).To(MatchError(ContainSubstring("Cannot specify both basic auth and UAA authentication")))
			})
		})

		Context("free flag is not specified", func() {
			BeforeEach(func() {
				configFileName = "config_with_no_free_flag.yml"
			})

			It("returns no error", func() {
				Expect(parseErr).NotTo(HaveOccurred())
			})

			It("parses free as nil", func() {
				Expect(conf.ServiceCatalog.Plans[0].Free).To(BeNil())
			})

		})
		Context("free flag is false", func() {
			BeforeEach(func() {
				configFileName = "config_with_free_false.yml"
			})

			It("returns no error", func() {
				Expect(parseErr).NotTo(HaveOccurred())
			})

			It("parses free as false", func() {
				Expect(*conf.ServiceCatalog.Plans[0].Free).To(BeFalse())
			})
		})
	})

	Context("when the file does not exist", func() {
		BeforeEach(func() {
			configFileName = "idontexist"
		})

		It("returns an error", func() {
			Expect(os.IsNotExist(parseErr)).To(BeTrue())
		})
	})

	Context("when the file does not contain valid YAML", func() {
		BeforeEach(func() {
			configFileName = "not_yaml.yml"
		})

		It("returns an error", func() {
			Expect(parseErr).To(MatchError(ContainSubstring("cannot unmarshal")))
		})
	})

	Describe("serializing to yaml", func() {
		It("doesn't serialize the persistent_disk plan field, if not specified", func() {
			conf := config.Config{
				ServiceCatalog: broker.ServiceCatalog{
					Plans: []broker.Plan{
						{
							ID:          "optional-disk-plan-id",
							Name:        "optional-disk-plan",
							Description: "optional-disk-plan-description",
							InstanceGroups: []serviceadapter.InstanceGroup{
								{
									VMType: "optional-disk-vm",
								},
							},
						},
					},
				},
			}

			Expect(yaml.Marshal(conf)).NotTo(ContainSubstring("persistent_disk"))
		})

		It("doesn't serialize the metadata/bullets, if not specified", func() {
			conf := config.Config{
				ServiceCatalog: broker.ServiceCatalog{
					Plans: []broker.Plan{
						{
							Metadata: broker.PlanMetadata{
								Bullets: nil,
							},
						},
					},
				},
			}

			Expect(yaml.Marshal(conf)).NotTo(ContainSubstring("bullets"))
		})

		It("doesn't serialize the lifecycle, if not specified", func() {
			conf := config.Config{
				ServiceCatalog: broker.ServiceCatalog{
					Plans: []broker.Plan{
						{
							InstanceGroups: []serviceadapter.InstanceGroup{
								{
									Lifecycle: "",
								},
							},
						},
					},
				},
			}
			Expect(yaml.Marshal(conf)).NotTo(ContainSubstring("lifecycle"))
		})

		It("doesn't serialize the quotas, if not specified", func() {
			conf := config.Config{
				ServiceCatalog: broker.ServiceCatalog{
					Metadata: broker.ServiceMetadata{},
					Plans: []broker.Plan{
						{
							Quotas: nil,
						},
					},
				},
			}

			Expect(yaml.Marshal(conf)).NotTo(ContainSubstring("quotas"))
		})

		It("doesn't serialize the requires, if not specified", func() {
			conf := config.Config{
				ServiceCatalog: broker.ServiceCatalog{
					Requires: nil,
				},
			}

			Expect(yaml.Marshal(conf)).NotTo(ContainSubstring("requires"))
		})

		It("doesn't serialize the dashboard_client, if not specified", func() {
			conf := config.Config{
				ServiceCatalog: broker.ServiceCatalog{
					DashboardClient: nil,
				},
			}

			Expect(yaml.Marshal(conf)).NotTo(ContainSubstring("dashboard_client"))
		})
	})
})

func booleanPointer(val bool) *bool {
	return &val
}
