package external_test

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/external"
	"github.com/pivotal-cf/on-demand-service-broker/external/fake_command_runner"
)

var _ = Describe("external service adapter", func() {
	const externalBinPath = "/thing"
	const validManifestContent = "name: a-service-deployment"

	var (
		a         *external.Adapter
		cmdRunner *fake_command_runner.FakeCommandRunner
		logs      *bytes.Buffer
	)

	BeforeEach(func() {
		logs = new(bytes.Buffer)
		cmdRunner = new(fake_command_runner.FakeCommandRunner)
		a = &external.Adapter{
			CommandRunner:   cmdRunner,
			ExternalBinPath: externalBinPath,
			Logger:          log.New(io.MultiWriter(GinkgoWriter, logs), "[unit-tests] ", log.LstdFlags),
		}
	})

	Describe("generating manifest", func() {
		var (
			serviceDeployment serviceadapter.ServiceDeployment
			plan              serviceadapter.Plan
			previousPlan      *serviceadapter.Plan
			params            map[string]interface{}
			previousManifest  []byte

			manifest    []byte
			generateErr error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(validManifestContent), []byte(""), intPtr(0), nil)

			serviceDeployment = serviceadapter.ServiceDeployment{
				DeploymentName: "a-service-deployment",
				Releases: serviceadapter.ServiceReleases{
					{Name: "a-bosh-release"},
				},
				Stemcell: serviceadapter.Stemcell{
					OS:      "BeOS",
					Version: "2",
				},
			}

			plan = serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"foo": "bar",
					"baz": map[interface{}]interface{}{
						"qux": "quux",
					},
				},
			}
			params = map[string]interface{}{
				"key": "value",
				"anotherkey": map[string]interface{}{
					"innerkey": "innervalue",
				},
			}
			previousManifest = []byte("a-manifest")
			previousPlan = &serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"previous": "props",
					"baz": map[interface{}]interface{}{
						"qux": "quux",
					},
				},
			}
		})

		JustBeforeEach(func() {
			manifest, generateErr = a.GenerateManifest(serviceDeployment, plan, params, previousManifest, previousPlan)
		})

		It("invokes external manifest generator with serialised parameters", func() {
			serialisedServiceDeployment, err := json.Marshal(serviceDeployment)
			Expect(err).NotTo(HaveOccurred())

			plan.Properties = external.SanitiseForJSON(plan.Properties)
			serialisedPlan, err := json.Marshal(plan)
			Expect(err).NotTo(HaveOccurred())

			serialisedParams, err := json.Marshal(params)
			Expect(err).NotTo(HaveOccurred())

			serialisedPreviousPlan, err := json.Marshal(previousPlan)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0)
			Expect(argsPassed).To(ConsistOf(externalBinPath, "generate-manifest",
				string(serialisedServiceDeployment), string(serialisedPlan),
				string(serialisedParams), string(previousManifest), string(serialisedPreviousPlan)))
		})

		Context("when the external service adapter succeeds", func() {
			Context("when the generated manifest is valid", func() {
				It("returns no error", func() {
					Expect(generateErr).ToNot(HaveOccurred())
				})

				It("returns the deserialised stdout from the external process as a bosh manifest", func() {
					Expect(manifest).To(Equal([]byte(validManifestContent)))
				})
			})

			Context("when the generated manifest is invalid", func() {
				Context("that has an incorrect deployment name", func() {
					BeforeEach(func() {
						invalidManifestContent := "name: not-the-deployment-name-given-to-the-adapter"
						cmdRunner.RunReturns([]byte(invalidManifestContent), []byte(""), intPtr(0), nil)
					})

					It("returns an error", func() {
						Expect(generateErr).To(MatchError("external service adapter generated manifest with an incorrect deployment name at /thing: Stdout: 'name: not-the-deployment-name-given-to-the-adapter', stderr: ''"))
					})

					It("logs that the external service adapter has generated a manifest with an incorrect deployment name", func() {
						Expect(logs).To(ContainSubstring("service adapter generated a manifest with an incorrect deployment name"))
					})
				})

				Context("that cannot be unmarshalled", func() {
					BeforeEach(func() {
						cmdRunner.RunReturns([]byte("unparseable"), []byte(""), intPtr(0), nil)
					})

					It("returns an error", func() {
						Expect(generateErr).To(MatchError("external service adapter generated manifest that is not valid YAML at /thing: Stdout: 'unparseable', stderr: ''"))
					})
				})
			})
		})

		Context("when the external service adapter exits with status 10", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(10), nil)
			})

			It("returns an error", func() {
				Expect(generateErr).To(MatchError("external service adapter exited with 10 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})

			It("logs that the external service adapter has not implemented the generate-manifest command", func() {
				Expect(logs).To(ContainSubstring("service adapter has not implemented generate-manifest command"))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(1), nil)
			})

			It("returns an error", func() {
				Expect(generateErr).To(MatchError("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("when the external service adapter fails, without an exit code", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(generateErr).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("previous plan is nil", func() {
			BeforeEach(func() {
				previousPlan = nil
			})

			It("returns no error", func() {
				Expect(generateErr).ToNot(HaveOccurred())
			})

			It("it writes 'null' to the argument list", func() {
				argsPassed := cmdRunner.RunArgsForCall(0)
				Expect(argsPassed[6]).To(Equal("null"))
			})
		})
	})

	Describe("creating binding", func() {

		var (
			bindingID          string
			deploymentTopology bosh.BoshVMs
			manifest           []byte
			arbitraryParams    map[string]interface{}

			adapterBinding   serviceadapter.Binding
			createBindingErr error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(`{
				"credentials": {
					"username": "user1",
					"password": "reallysecret"
				},
				"syslog_drain_url": "some-url",
				"route_service_url": "route-url"
			}`), []byte(""), intPtr(0), nil)

			bindingID = "the-binding"
			deploymentTopology = bosh.BoshVMs{"the-deployment": []string{"a-vm"}}
			manifest = []byte("a-manifest")
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})

		JustBeforeEach(func() {
			adapterBinding, createBindingErr = a.CreateBinding(bindingID, deploymentTopology, manifest, arbitraryParams)
		})

		It("invokes external binding creator with serialised params", func() {
			serialisedVMs, err := json.Marshal(deploymentTopology)
			Expect(err).NotTo(HaveOccurred())

			serialisedArbitraryParams, err := json.Marshal(arbitraryParams)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0)
			Expect(argsPassed).To(ConsistOf(externalBinPath, "create-binding", bindingID, string(serialisedVMs), string(manifest), string(serialisedArbitraryParams)))
		})

		Context("when the external adapter succeeds", func() {
			It("returns no error", func() {
				Expect(createBindingErr).ToNot(HaveOccurred())
			})

			It("returns the service-specific binding output", func() {
				Expect(adapterBinding).To(Equal(
					serviceadapter.Binding{
						Credentials:     map[string]interface{}{"username": "user1", "password": "reallysecret"},
						SyslogDrainURL:  "some-url",
						RouteServiceURL: "route-url",
					},
				))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(1), nil)
			})

			It("returns an error", func() {
				Expect(createBindingErr).To(MatchError("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("when the external adapter fails with exit code 49", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(49), nil)
			})

			It("returns an error", func() {
				Expect(createBindingErr).To(Equal(serviceadapter.NewBindingAlreadyExistsError(fmt.Errorf("external service adapter exited with 49 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))))
			})
		})

		Context("when the external adapter fails with exit code 42", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(42), nil)
			})

			It("returns an error", func() {
				Expect(createBindingErr).To(Equal(serviceadapter.NewAppGuidNotProvidedError(fmt.Errorf("external service adapter exited with 42 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))))
			})
		})

		Context("when the external adapter fails with exit code 10", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(10), nil)
			})
			It("returns an error", func() {
				Expect(createBindingErr).To(Equal(fmt.Errorf("external service adapter exited with 10 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
			})
			It("logs that the external service adapter has not implemented the create-binding command", func() {
				Expect(logs).To(ContainSubstring("service adapter has not implemented create-binding command"))
			})
		})

		Context("when the external adapter fails with no exit code", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
			})
			It("returns an error", func() {
				Expect(createBindingErr).To(Equal(fmt.Errorf("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
			})
		})
	})

	Describe("deleting binding", func() {
		var (
			bindingID          string
			deploymentTopology bosh.BoshVMs
			manifest           []byte

			deleteBindingError error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(""), []byte(""), intPtr(0), nil)

			bindingID = "the-binding"
			deploymentTopology = bosh.BoshVMs{}
			manifest = []byte("a-manifest")
		})

		JustBeforeEach(func() {
			deleteBindingError = a.DeleteBinding(bindingID, deploymentTopology, manifest)
		})

		It("invokes external binary with params to delete binding", func() {
			serialisedBoshVMs, err := json.Marshal(deploymentTopology)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0)
			Expect(argsPassed).To(ConsistOf(externalBinPath, "delete-binding", bindingID, string(serialisedBoshVMs), string(manifest)))
		})

		Context("when the external adapter succeeds", func() {
			It("returns no error", func() {
				Expect(deleteBindingError).ToNot(HaveOccurred())
			})
		})

		Context("when the external adapter fails with no exit code", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
			})
			It("returns an error", func() {
				Expect(deleteBindingError).To(Equal(fmt.Errorf("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
			})
		})

		Context("when the external adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(1), nil)
			})

			It("returns an error", func() {
				Expect(deleteBindingError).To(MatchError("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("when the external adapter fails with exit code 10", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(10), nil)
			})

			It("returns an error", func() {
				Expect(deleteBindingError).To(Equal(fmt.Errorf("external service adapter exited with 10 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'")))
			})
			It("logs that the external service adapter has not implemented the delete-binding command", func() {
				Expect(logs).To(ContainSubstring("service adapter has not implemented delete-binding command"))
			})
		})
	})

	Describe("dashboard url", func() {
		var (
			instanceID string
			plan       serviceadapter.Plan
			manifest   []byte

			actualDashboardUrl string
			actualError        error
		)

		BeforeEach(func() {
			plan = serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"foo": "bar",
					"baz": map[string]interface{}{
						"qux": "quux",
					},
				},
			}

			manifest = []byte("a manifest")

			cmdRunner.RunReturns([]byte(""), []byte(""), intPtr(0), nil)

			instanceID = "the-instance-id"
			manifest = []byte("a-manifest")

			dashboardUrlJSON := []byte(`{ "dashboard_url": "https://someurl.com"}`)

			cmdRunner.RunReturns(dashboardUrlJSON, []byte(""), intPtr(0), nil)
		})

		JustBeforeEach(func() {
			actualDashboardUrl, actualError = a.GenerateDashboardUrl(instanceID, plan, manifest)
		})

		It("invokes external dashboard url generator with serialised params", func() {
			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			planJson, err := json.Marshal(plan)
			Expect(err).NotTo(HaveOccurred())
			argsPassed := cmdRunner.RunArgsForCall(0)
			Expect(argsPassed).To(ConsistOf(externalBinPath, "dashboard-url", instanceID, string(planJson), string(manifest)))
		})

		It("returns a dashboard url", func() {
			Expect(actualDashboardUrl).To(Equal("https://someurl.com"))
		})

		It("has no error", func() {
			Expect(actualError).NotTo(HaveOccurred())
		})

		Context("when plan properties are formatted as map[interface][interface]", func() {
			BeforeEach(func() {
				plan = serviceadapter.Plan{
					Properties: serviceadapter.Properties{
						"foo": "bar",
						"baz": map[interface{}]interface{}{
							"qux": "quux",
						},
					},
				}
			})
			It("converts plan properties to be json serializable", func() {
				Expect(actualError).NotTo(HaveOccurred())
				Expect(cmdRunner.RunCallCount()).To(Equal(1))
				argsPassed := cmdRunner.RunArgsForCall(0)

				convertedPlan := serviceadapter.Plan{
					Properties: serviceadapter.Properties{
						"foo": "bar",
						"baz": map[string]interface{}{
							"qux": "quux",
						},
					},
				}

				convertedPlanJson, err := json.Marshal(convertedPlan)
				Expect(err).NotTo(HaveOccurred())
				Expect(argsPassed).To(ConsistOf(externalBinPath, "dashboard-url", instanceID, string(convertedPlanJson), string(manifest)))
			})
		})

		Context("adapter did not implement the dashboard url subcommand", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte(""), []byte("not impemented"), intPtr(10), nil)
			})
			It("returns the subcommand not implemented error", func() {
				Expect(actualError).To(HaveOccurred())
				Expect(actualError).To(BeAssignableToTypeOf(broker.NewServiceAdapterCommandNotImplementedError(fmt.Errorf("no relvent"))))
			})
		})

		Context("when the external service adapter fails, without an exit code", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), nil, errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(actualError).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), intPtr(1), nil)
			})

			It("returns an error", func() {
				Expect(actualError).To(MatchError("external service adapter exited with 1 at /thing: Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("cannot deserailse response", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("invalid json"), []byte("I'm stderr"), intPtr(0), nil)
			})

			It("returns an error", func() {
				Expect(actualError).To(HaveOccurred())
			})
		})
	})
})

func intPtr(i int) *int {
	return &i
}
