package instance_repository_test

import (
	"database/sql"
	"os"
	"os/exec"
	"path/filepath"

	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/instance_repository"

	_ "github.com/mattn/go-sqlite3"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("InstanceRepository", func() {
	var (
		dbConn       *sql.DB
		dbPath       string
		instanceRepo *instance_repository.SQLiteInstanceRepository
	)

	BeforeEach(func() {
		cwd, err := os.Getwd()
		Expect(err).NotTo(HaveOccurred())
		dbPath = filepath.Join(cwd, "test_db", "store.db")

		dbConn, err = sql.Open("sqlite3", dbPath)
		Expect(err).NotTo(HaveOccurred())

		gooseUp := exec.Command("goose", "-path", ".", "up")
		gooseUp.Stdout = GinkgoWriter
		gooseUp.Stderr = GinkgoWriter
		Expect(gooseUp.Run()).To(Succeed())

		instanceRepo = instance_repository.New(dbConn)
	})

	AfterEach(func() {
		Expect(dbConn.Close()).To(Succeed())
		Expect(os.Remove(dbPath)).To(Succeed())
	})

	Describe("saving instances", func() {
		var (
			planID     = "plan"
			instanceID = "some-instance"

			instance broker.Instance
			saveErr  error
		)

		BeforeEach(func() {
			instance = broker.Instance{ID: instanceID, PlanID: planID, LastOperationType: broker.Create}
		})

		JustBeforeEach(func() {
			saveErr = instanceRepo.Save(instance)
		})

		It("does not error", func() {
			Expect(saveErr).NotTo(HaveOccurred())
		})

		Context("when the instance is updated and saved again", func() {
			var (
				updateErr error
			)

			JustBeforeEach(func() {
				instance.LastOperationType = broker.Update
				updateErr = instanceRepo.Save(instance)
			})

			It("does not error", func() {
				Expect(updateErr).NotTo(HaveOccurred())
			})

			Context("and the instance is retrieved", func() {
				It("has the new task ID", func() {
					i, err := instanceRepo.FindByID(instance.ID)
					Expect(err).NotTo(HaveOccurred())
					Expect(i.LastOperationType).To(Equal(broker.Update))
				})
			})
		})

		Describe("retrieving instance by ID", func() {
			var (
				instanceIDToRetrieve string

				retrievedInstance = broker.Instance{ID: "some-instance"}
				retrieveErr       error
			)

			JustBeforeEach(func() {
				retrievedInstance, retrieveErr = instanceRepo.FindByID(instanceIDToRetrieve)
			})

			Context("when the instance exists", func() {
				BeforeEach(func() {
					instanceIDToRetrieve = instance.ID
				})

				It("does not error", func() {
					Expect(retrieveErr).NotTo(HaveOccurred())
				})

				It("retrieves the instance", func() {
					Expect(retrievedInstance).To(Equal(instance))
				})
			})

			Context("when the instance does not exist", func() {
				BeforeEach(func() {
					instanceIDToRetrieve = "idontexist"
				})

				It("returns an error", func() {
					Expect(retrieveErr).To(MatchError(sql.ErrNoRows))
				})
			})
		})

		saveAnInstanceWithIDs := func(id string, planID string) {
			err := instanceRepo.Save(broker.Instance{ID: id, PlanID: planID})
			Expect(err).NotTo(HaveOccurred())
		}

		Describe("check if instance with id exists", func() {
			It("returns true if instance exists", func() {
				Expect(instanceRepo.InstanceExists(instanceID)).To(BeTrue())
			})
			It("returns false if instance does not exists", func() {
				Expect(instanceRepo.InstanceExists("non-exittant")).To(BeFalse())
			})
		})

		Describe("counting instances with a particular plan ID", func() {
			It("counts an instance with a matching plan ID", func() {
				Expect(instanceRepo.CountByPlanID(planID)).To(Equal(1))
			})

			It("counts an instance with a matching plan ID once after an update", func() {
				saveAnInstanceWithIDs(instanceID, planID)
				Expect(instanceRepo.CountByPlanID(planID)).To(Equal(1))
			})

			It("counts all instances with matching plan IDs", func() {
				saveAnInstanceWithIDs("something-else", planID)
				Expect(instanceRepo.CountByPlanID(planID)).To(Equal(2))
			})

			It("does not count an instance with a non-matching plan ID", func() {
				saveAnInstanceWithIDs("something-else", "something-wrong")
				Expect(instanceRepo.CountByPlanID(planID)).To(Equal(1))
			})
		})

		Describe("deleting an instance", func() {
			var (
				instanceToDelete string
				deleteErr        error
			)

			JustBeforeEach(func() {
				deleteErr = instanceRepo.DeleteByID(instanceToDelete)
			})

			Context("when the instance exists", func() {
				BeforeEach(func() {
					instanceToDelete = instance.ID
				})

				It("returns no error", func() {
					Expect(deleteErr).NotTo(HaveOccurred())
				})

				It("cannot subsequently be retrieved", func() {
					_, err := instanceRepo.FindByID(instanceToDelete)
					Expect(err).To(HaveOccurred())
				})
			})

			Context("when the instance does not exist", func() {
				BeforeEach(func() {
					instanceToDelete = "idontexist"
				})

				It("returns an error", func() {
					Expect(deleteErr).To(MatchError(ContainSubstring("error deleting instance idontexist")))
				})
			})
		})

		Describe("retrieving all instances", func() {
			It("returns all instances", func() {
				instances, err := instanceRepo.AllInstances()
				Expect(err).NotTo(HaveOccurred())
				Expect(instances).To(ConsistOf(instance))
			})
		})
	})
})
