package instance_repository

import (
	"database/sql"
	"fmt"

	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

type SQLiteInstanceRepository struct {
	conn *sql.DB
}

func New(dbConn *sql.DB) *SQLiteInstanceRepository {
	return &SQLiteInstanceRepository{
		conn: dbConn,
	}
}

func (r *SQLiteInstanceRepository) FindByID(instanceID string) (broker.Instance, error) {
	instance, err := scanInstanceRow(r.conn.QueryRow("SELECT * FROM instances WHERE id=?", instanceID))
	if err == sql.ErrNoRows {
		return broker.Instance{}, err
	} else if err != nil {
		return broker.Instance{}, fmt.Errorf("error retrieving instance %s: %s", instanceID, err)
	}
	return instance, nil
}

func (r *SQLiteInstanceRepository) Save(instance broker.Instance) error {
	_, err := r.conn.Exec("INSERT OR REPLACE INTO instances(id, planid, lastoperationtype) VALUES(?, ?, ?)", instance.ID, instance.PlanID, string(instance.LastOperationType))
	return err
}

func (r *SQLiteInstanceRepository) InstanceExists(instanceID string) (bool, error) {
	_, err := r.FindByID(instanceID)
	switch err {
	case sql.ErrNoRows:
		return false, nil
	case nil:
		return true, nil
	default:
		return false, err
	}
	return true, nil
}

func (r *SQLiteInstanceRepository) AllInstances() ([]broker.Instance, error) {
	rows, err := r.conn.Query("SELECT * from instances")
	if err != nil {
		return []broker.Instance{}, nil
	}

	instances := []broker.Instance{}
	for rows.Next() {
		instance, err := scanInstanceRow(rows)
		if err != nil {
			return []broker.Instance{}, err
		}
		instances = append(instances, instance)
	}
	return instances, nil
}

func (r *SQLiteInstanceRepository) CountByPlanID(planID string) (int, error) {
	row := r.conn.QueryRow("SELECT count(*) from instances where planid = ?", planID)
	var count int = -1
	err := row.Scan(&count)
	return count, err
}

func (r *SQLiteInstanceRepository) DeleteByID(instanceID string) error {
	errs := func(err error) error {
		return fmt.Errorf("error deleting instance %s: %s", instanceID, err)
	}

	result, err := r.conn.Exec("DELETE FROM instances WHERE id=?", instanceID)
	if err != nil {
		return errs(err)
	}

	rowsAffected, err := result.RowsAffected()
	if err != nil {
		return errs(err)
	}

	if rowsAffected != 1 {
		return errs(fmt.Errorf("expected 1 row to be deleted, got %d", rowsAffected))
	}

	return nil
}

type scannable interface {
	Scan(...interface{}) error
}

func scanInstanceRow(row scannable) (broker.Instance, error) {
	var instance broker.Instance

	var lastOperationType string
	err := row.Scan(&instance.ID, &instance.PlanID, &lastOperationType)
	if err != nil {
		return instance, err
	}
	instance.LastOperationType = broker.OperationType(lastOperationType)
	return instance, nil
}
