package mgmtapi

import (
	"encoding/json"
	"io"
	"log"
	"net/http"

	"github.com/gorilla/mux"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

type api struct {
	manageableBroker ManageableBroker
	logger           *log.Logger
}

//go:generate counterfeiter -o fake_manageable_broker/fake_manageable_broker.go . ManageableBroker
type ManageableBroker interface {
	Instances() ([]string, error)
	Upgrade(instanceID string) error
}

func AttachRoutes(r *mux.Router, manageableBroker ManageableBroker, logger *log.Logger) {
	a := &api{manageableBroker: manageableBroker, logger: logger}
	r.HandleFunc("/mgmt/service_instances", a.listAllInstances).Methods("GET")
	r.HandleFunc("/mgmt/service_instances/{instance_id}", a.upgradeInstance).Methods("PATCH")
}

func (a *api) listAllInstances(w http.ResponseWriter, r *http.Request) {
	instanceIDs, err := a.manageableBroker.Instances()
	if err != nil {
		a.logger.Printf("error occurred querying instances: %s", err)
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	a.writeJson(w, instanceIDs)
}

func (a *api) upgradeInstance(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	instanceID := vars["instance_id"]
	err := a.manageableBroker.Upgrade(instanceID)
	switch err.(type) {
	case nil:
		w.WriteHeader(http.StatusAccepted)
	case broker.DeploymentNotFoundError:
		w.WriteHeader(http.StatusGone)
	case broker.DeploymentInProgressError:
		w.WriteHeader(http.StatusConflict)
	case error:
		a.logger.Printf("error occurred upgrading instance %s: %s", instanceID, err)
		w.WriteHeader(http.StatusInternalServerError)
		a.writeJson(w, brokerapi.ErrorResponse{Description: err.Error()})
	}
}

func (a *api) writeJson(w io.Writer, obj interface{}) {
	if err := json.NewEncoder(w).Encode(obj); err != nil {
		a.logger.Printf("error occurred encoding json: %s", err)
	}
}
