package mgmtapi_test

import (
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"net/http/httptest"

	"github.com/gorilla/mux"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi/fake_manageable_broker"
)

var _ = Describe("Management API", func() {
	var (
		server           *httptest.Server
		manageableBroker *fake_manageable_broker.FakeManageableBroker
		logger           = log.New(GinkgoWriter, "[mgmtapi unit tests] ", log.LstdFlags)
	)

	BeforeEach(func() {
		manageableBroker = new(fake_manageable_broker.FakeManageableBroker)

		router := mux.NewRouter()
		mgmtapi.AttachRoutes(router, manageableBroker, logger)
		server = httptest.NewServer(router)
	})

	AfterEach(func() {
		server.Close()
	})

	Describe("listing all instances", func() {
		var listResp *http.Response

		JustBeforeEach(func() {
			var err error
			listResp, err = http.Get(fmt.Sprintf("%s/mgmt/service_instances", server.URL))
			Expect(err).NotTo(HaveOccurred())
		})

		Context("of which there are three", func() {
			BeforeEach(func() {
				instances := []string{"red", "green", "blue"}
				manageableBroker.InstancesReturns(instances, nil)
			})

			It("returns HTTP 200", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusOK))
			})

			It("returns an empty JSON array", func() {
				var instances []string
				Expect(json.NewDecoder(listResp.Body).Decode(&instances)).To(Succeed())
				Expect(instances).To(ConsistOf("red", "green", "blue"))
			})
		})

		Context("but failing to do so", func() {
			BeforeEach(func() {
				manageableBroker.InstancesReturns(nil, errors.New("fnuh"))
			})

			It("returns HTTP 500", func() {
				Expect(listResp.StatusCode).To(Equal(http.StatusInternalServerError))
			})
		})
	})

	Describe("upgrading an instance", func() {
		var upgradeResp *http.Response
		var instanceID = "283974"

		JustBeforeEach(func() {
			var err error
			upgradeResp, err = Patch(fmt.Sprintf("%s/mgmt/service_instances/"+instanceID, server.URL))
			Expect(err).NotTo(HaveOccurred())
		})

		Context("successfully", func() {
			It("upgrades the instance using the broker", func() {
				Expect(manageableBroker.UpgradeCallCount()).To(Equal(1))
				Expect(manageableBroker.UpgradeArgsForCall(0)).To(Equal(instanceID))
			})

			It("returns HTTP 202", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusAccepted))
			})
		})

		Context("there is a skipable error", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(broker.NewDeploymentNotFoundError(fmt.Errorf("this did not work")))
			})

			It("returns HTTP 410", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusGone))
			})
		})

		Context("there is a deployment in progress error", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(broker.NewDeploymentInProgressError(fmt.Errorf("this did not work")))
			})

			It("returns HTTP 409", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusConflict))
			})
		})

		Context("unsuccessfully", func() {
			BeforeEach(func() {
				manageableBroker.UpgradeReturns(fmt.Errorf("this did not work"))
			})

			It("returns HTTP 500", func() {
				Expect(upgradeResp.StatusCode).To(Equal(http.StatusInternalServerError))
			})

			It("returns the error message", func() {
				Expect(ioutil.ReadAll(upgradeResp.Body)).To(MatchJSON(`{"description": "this did not work"}`))
			})
		})
	})
})

func Patch(url string) (resp *http.Response, err error) {
	req, err := http.NewRequest("PATCH", url, nil)
	if err != nil {
		return nil, err
	}
	return http.DefaultClient.Do(req)
}
