package upgrader_test

import (
	"net/http"
	"strings"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/onsi/gomega/ghttp"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/upgrader"
)

var _ = Describe("Upgrader", func() {
	var server *ghttp.Server

	var brokerUsername, brokerPassword, brokerUrl string
	var pollingInterval int64

	var successfulInstances, skippedInstances []string
	var actualErr error

	BeforeEach(func() {
		brokerUsername = "foo"
		brokerPassword = "bar"
		pollingInterval = 1
		server = ghttp.NewServer()
		brokerUrl = server.URL()
	})

	AfterEach(func() {
		server.Close()
	})

	JustBeforeEach(func() {
		successfulInstances, skippedInstances, actualErr = upgrader.New(brokerUsername, brokerPassword, brokerUrl, pollingInterval).Upgrade()
	})

	Describe("authentication", func() {
		Context("success", func() {
			BeforeEach(func() {
				server.AppendHandlers(ghttp.CombineHandlers(
					ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
					ghttp.RespondWithJSONEncoded(http.StatusOK, []string{}),
					ghttp.VerifyBasicAuth(brokerUsername, brokerPassword),
				))
			})

			It("autenticates with the given credentials", func() {
				Expect(successfulInstances).To(BeEmpty())
				Expect(actualErr).NotTo(HaveOccurred())
			})
		})

		Context("failure", func() {
			BeforeEach(func() {
				server.AppendHandlers(ghttp.CombineHandlers(
					ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
					ghttp.RespondWithJSONEncoded(http.StatusUnauthorized, []string{}),
				))
			})

			It("returns an error", func() {
				Expect(actualErr).To(HaveOccurred())
			})
		})
	})

	Context("upgrades one instance, successfully", func() {
		serviceInstanceId := "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstanceId}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstanceId+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})

		It("returns the list of successful upgrades", func() {
			Expect(successfulInstances).To(Equal([]string{serviceInstanceId}))
			Expect(actualErr).NotTo(HaveOccurred())
			Expect(skippedInstances).To(BeEmpty())
		})
	})

	Context("upgrades one instance, calls last operation till successful", func() {
		serviceInstanceId := "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstanceId}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstanceId+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.InProgress}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstanceId+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.InProgress}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstanceId+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})

		It("returns the list of successful upgrades", func() {
			Expect(successfulInstances).To(Equal([]string{serviceInstanceId}))
			Expect(actualErr).NotTo(HaveOccurred())
		})

		It("checks last_operation api 3 times", func() {
			calls := 0
			for _, req := range server.ReceivedRequests() {
				if strings.HasSuffix(req.URL.String(), "/last_operation") {
					calls++
				}
			}
			Expect(calls).To(Equal(3))
		})
	})

	Context("skips one instance, if gone", func() {
		serviceInstanceId := "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstanceId}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusGone, nil),
			))
		})

		It("returns the skipped instance", func() {
			Expect(successfulInstances).To(BeEmpty())
			Expect(skippedInstances).To(Equal([]string{serviceInstanceId}))
			Expect(actualErr).NotTo(HaveOccurred())
		})
	})

	Context("retires upgrade if conflict, till accepted", func() {
		serviceInstanceId := "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstanceId}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstanceId+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})

		It("returns the skipped instance", func() {
			Expect(successfulInstances).To(Equal([]string{serviceInstanceId}))
			Expect(skippedInstances).To(BeEmpty())
			Expect(actualErr).NotTo(HaveOccurred())
		})

		It("it tries upgrade 4 times", func() {
			calls := 0
			for _, req := range server.ReceivedRequests() {
				if strings.HasSuffix(req.URL.String(), "/mgmt/service_instances/"+serviceInstanceId) {
					calls++
				}
			}
			Expect(calls).To(Equal(4))
		})
	})

	Context("retires upgrade if conflict, till skipped", func() {
		serviceInstanceId := "serviceInstanceId"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstanceId}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstanceId),
				ghttp.RespondWithJSONEncoded(http.StatusGone, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstanceId+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})

		It("returns the skipped instance", func() {
			Expect(successfulInstances).To(BeEmpty())
			Expect(skippedInstances).To(Equal([]string{serviceInstanceId}))
			Expect(actualErr).NotTo(HaveOccurred())
		})

		It("it tries upgrade 4 times", func() {
			calls := 0
			for _, req := range server.ReceivedRequests() {
				if strings.HasSuffix(req.URL.String(), "/mgmt/service_instances/"+serviceInstanceId) {
					calls++
				}
			}
			Expect(calls).To(Equal(4))
		})
	})

	Context("upgrades multiple instances, successfully", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstance1, serviceInstance2, serviceInstance3}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance1),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance1+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance2+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance3),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance3+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})

		It("returns the list of successful upgrades", func() {
			Expect(successfulInstances).To(Equal([]string{serviceInstance1, serviceInstance2, serviceInstance3}))
			Expect(actualErr).NotTo(HaveOccurred())
		})
	})

	Context("upgrades multiple instances, upgrade fails", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		expectedErrorMessage := "everything went wrong"
		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstance1, serviceInstance2, serviceInstance3}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance1),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance1+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusInternalServerError, brokerapi.ErrorResponse{Description: expectedErrorMessage}),
			))
		})
		It("returns the error with the description from the http response", func() {
			Expect(actualErr).To(MatchError(ContainSubstring(expectedErrorMessage)))
		})
	})

	Context("upgrades multiple instances, last operation fails", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"
		expectedErrorMessage := "everything went wrong"

		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstance1, serviceInstance2, serviceInstance3}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance1),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance1+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance2+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Failed, Description: expectedErrorMessage}),
			))
		})

		It("returns the error with the description from the http response", func() {
			Expect(actualErr).To(MatchError(ContainSubstring(expectedErrorMessage)))
		})
	})

	Context("upgrades multiple instances, with one instance skipped", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"

		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstance1, serviceInstance2, serviceInstance3}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance1),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance1+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusGone, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance3),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance3+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})
		It("returns successfully", func() {
			Expect(actualErr).NotTo(HaveOccurred())
		})

		It("returns a list of skipped instances", func() {
			Expect(skippedInstances).To(ConsistOf(serviceInstance2))
		})

		It("returns a list of upgraded instances", func() {
			Expect(successfulInstances).To(ConsistOf(serviceInstance1, serviceInstance3))
		})
	})

	Context("upgrades multiple instances, by retrying instances which where in progress", func() {
		serviceInstance1 := "serviceInstanceId1"
		serviceInstance2 := "serviceInstanceId2"
		serviceInstance3 := "serviceInstanceId3"

		BeforeEach(func() {
			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/mgmt/service_instances"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, []string{serviceInstance1, serviceInstance2, serviceInstance3}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance1),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance1+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance3),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance3+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusConflict, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("PATCH", "/mgmt/service_instances/"+serviceInstance2),
				ghttp.RespondWithJSONEncoded(http.StatusAccepted, nil),
			))

			server.AppendHandlers(ghttp.CombineHandlers(
				ghttp.VerifyRequest("GET", "/v2/service_instances/"+serviceInstance2+"/last_operation"),
				ghttp.RespondWithJSONEncoded(http.StatusOK, brokerapi.LastOperation{State: brokerapi.Succeeded}),
			))
		})

		It("returns successfully", func() {
			Expect(actualErr).NotTo(HaveOccurred())
		})

		It("skipped no instances", func() {
			Expect(skippedInstances).To(BeEmpty())
		})

		It("upgraded all instances", func() {
			Expect(successfulInstances).To(ConsistOf(serviceInstance1, serviceInstance3, serviceInstance2))
		})
	})

})
